"""@file galaxycluster.py
The GalaxyCluster class
"""

import pickle
import warnings

from .dataops import (
    compute_background_probability,
    compute_galaxy_weights,
    compute_tangential_and_cross_components,
    make_radial_profile,
)
from .gcdata import GCData
from .theory import compute_critical_surface_density_eff
from .utils import (
    _draw_random_points_from_tab_distribution,
    _validate_dec,
    _validate_ra,
    validate_argument,
)
from .utils.plotting import plot_profiles


class GalaxyCluster:
    """Object that contains the galaxy cluster metadata and background galaxy data

    Attributes
    ----------
    unique_id : int or string
        Unique identifier of the galaxy cluster
    ra : float
        Right ascension of galaxy cluster center (in degrees)
    dec : float
        Declination of galaxy cluster center (in degrees)
    z : float
        Redshift of galaxy cluster center
    galcat : GCData
        Table of background galaxy data containing at least galaxy_id, ra, dec, e1, e2, z
    validate_input: bool
        Validade each input argument
    """

    def __init__(self, *args, validate_input=True, **kwargs):
        self.unique_id = None
        self.ra = None
        self.dec = None
        self.z = None
        self.galcat = None
        self.validate_input = validate_input
        if len(args) > 0 or len(kwargs) > 0:
            self._add_values(*args, **kwargs)
            self._check_types()
            self.set_ra_lower(ra_low=0)

    def _add_values(
        self,
        unique_id: str,
        ra: float,
        dec: float,
        z: float,
        galcat: GCData,
    ):
        """Add values for all attributes"""
        self.unique_id = unique_id
        self.ra = ra
        self.dec = dec
        self.z = z
        self.galcat = galcat

    def _check_types(self):
        """Check types of all attributes"""
        validate_argument(vars(self), "unique_id", (int, str))
        _validate_ra(vars(self), "ra", False)
        _validate_dec(vars(self), "dec", False)
        validate_argument(vars(self), "z", (float, str), argmin=0, eqmin=True)
        validate_argument(vars(self), "galcat", GCData)
        self.unique_id = str(self.unique_id)
        self.ra = float(self.ra)
        self.dec = float(self.dec)
        self.z = float(self.z)

    def save(self, filename, **kwargs):
        """Saves GalaxyCluster object to filename using Pickle"""
        with open(filename, "wb") as fin:
            pickle.dump(self, fin, **kwargs)

    @classmethod
    def load(cls, filename, **kwargs):
        """Loads GalaxyCluster object to filename using Pickle"""
        with open(filename, "rb") as fin:
            self = pickle.load(fin, **kwargs)
        # pylint: disable=protected-access
        self._check_types()
        return self

    def _str_colnames(self):
        """Colnames in comma separated str"""
        return ", ".join(self.galcat.colnames)

    def __repr__(self):
        """Generates basic description of GalaxyCluster"""
        return (
            f"GalaxyCluster {self.unique_id}: "
            f"(ra={self.ra}, dec={self.dec}) at z={self.z}"
            f"\n> with columns: {self._str_colnames()}"
            f"\n> {len(self.galcat)} source galaxies"
        )

    def __str__(self):
        """Generates string for print(GalaxyCluster)"""
        table = "objects".join(self.galcat.__str__().split("objects")[1:])
        return self.__repr__() + "\n" + table

    def _repr_html_(self):
        """Generates string for display(GalaxyCluster)"""
        # pylint: disable=protected-access
        return (
            f"<b>GalaxyCluster:</b> {self.unique_id} "
            f"(ra={self.ra}, dec={self.dec}) at z={self.z}"
            f"<br>> <b>with columns:</b> {self._str_colnames()}"
            f"<br>> {len(self.galcat)} source galaxies"
            f"<br>{self.galcat._html_table()}"
        )

    def add_critical_surface_density(self, cosmo, use_pdz=False, force=False):
        r"""Computes the critical surface density for each galaxy in `galcat`.
        It only runs if input cosmo != galcat cosmo or if `sigma_c` not in `galcat`.

        Parameters
        ----------
        cosmo : clmm.Cosmology object
            CLMM Cosmology object
        use_pdz : bool
            Flag to specify the use of the photoz pdf. If `False` (default), `sigma_c` is computed
            using the redshift point estimates of the `z` column of the `galcat` table. If `True`,
            `sigma_c` is computed as 1/<1/Sigma_crit>, where the average is performed using
            the individual galaxy redshift pdf. In that case, the `galcat` table should have
            pzbins` and `pzpdf` columns.
        force : bool
            Force recomputation of sigma_c.

        Returns
        -------
        None
        """
        if cosmo is None:
            raise TypeError("To compute Sigma_crit, please provide a cosmology")
        sigmac_colname = "sigma_c_eff" if use_pdz else "sigma_c"
        if (
            cosmo.get_desc() != self.galcat.meta["cosmo"]
            or sigmac_colname not in self.galcat.columns
            or force
        ):
            if self.z is None:
                raise TypeError("Cluster's redshift is None. Cannot compute Sigma_crit")
            if not use_pdz and "z" not in self.galcat.columns:
                raise TypeError(
                    "Galaxy catalog missing the redshift column (which should be"
                    "called 'z'). Cannot compute Sigma_crit."
                )
            if use_pdz and not self.galcat.has_pzpdfs():
                raise TypeError(
                    "Galaxy catalog missing the pzpdfs. Cannot compute 1/<1/Sigma_crit>."
                )

            self.galcat.update_cosmo(cosmo, overwrite=True)
            if not use_pdz:
                self.galcat[sigmac_colname] = cosmo.eval_sigma_crit(self.z, self.galcat["z"])
            else:
                zdata = self._get_input_galdata({"pzpdf": "pzpdf", "pzbins": "pzbins"})
                self.galcat[sigmac_colname] = compute_critical_surface_density_eff(
                    cosmo=cosmo,
                    z_cluster=self.z,
                    pzbins=zdata["pzbins"],
                    pzpdf=zdata["pzpdf"],
                    validate_input=self.validate_input,
                )
        return sigmac_colname

    def _get_input_galdata(self, col_dict):
        """
        Checks required columns exist in galcat and returns kwargs dictionary
        to be passed to dataops functions.

        Parametters
        -----------
        col_dict : dict
            Dictionary with the names of the dataops arguments as keys and galcat columns
            as values, made to usually pass locals() here.

        Returns
        -------
        kwarg_data : dict
            Dictionary with the data to be passed to functions by **kwargs method.
        """
        use_cols = {**col_dict}
        kwarg_data = {}
        if "pzbins" in col_dict:
            if not self.galcat.has_pzpdfs():
                raise TypeError("Missing galaxy photoz distributions")
            use_cols.pop("pzbins")
            use_cols.pop("pzpdf")
            pzbins, pzpdfs = self.galcat.get_pzpdfs()
            kwarg_data.update({"pzbins": pzbins, "pzpdf": pzpdfs})
        missing_cols = ", ".join(
            [f"'{t_}'" for t_ in use_cols.values() if t_ not in self.galcat.columns]
        )
        if len(missing_cols) > 0:
            raise TypeError(f"Galaxy catalog missing required columns: {missing_cols}")
        kwarg_data.update({key: self.galcat[colname] for key, colname in use_cols.items()})
        return kwarg_data

    def compute_tangential_and_cross_components(
        self,
        shape_component1="e1",
        shape_component2="e2",
        tan_component="et",
        cross_component="ex",
        geometry="curve",
        is_deltasigma=False,
        use_pdz=False,
        cosmo=None,
        add=True,
    ):
        r"""Adds a tangential- and cross- components for shear or ellipticity to self

        Calls `clmm.dataops.compute_tangential_and_cross_components` with the following arguments:
        ra_lens: `cluster` Ra
        dec_lens: `cluster` Dec
        ra_source: `galcat` Ra
        dec_source: `galcat` Dec
        shear1: `galcat` shape_component1
        shear2: `galcat` shape_component2
        geometry: `input` geometry
        is_deltasigma: `input` is_deltasigma
        coordinate_system: `galcat` coordinate_system

        Parameters
        ----------
        shape_component1: string, optional
            Name of the column in the `galcat` astropy table of the cluster object that contains
            the shape or shear measurement along the first axis. Default: `e1`
        shape_component1: string, optional
            Name of the column in the `galcat` astropy table of the cluster object that contains
            the shape or shear measurement along the second axis. Default: `e2`
        tan_component: string, optional
            Name of the column to be added to the `galcat` astropy table that will contain the
            tangential component computed from columns `shape_component1` and `shape_component2`.
            Default: `et`
        cross_component: string, optional
            Name of the column to be added to the `galcat` astropy table that will contain the
            cross component computed from columns `shape_component1` and `shape_component2`.
            Default: `ex`
        geometry: str, optional
            Sky geometry to compute angular separation.
            Options are curve (uses astropy) or flat.
        is_deltasigma: bool
            If `True`, the tangential and cross components returned are multiplied by Sigma_crit.
            Results in units of :math:`M_\odot\ Mpc^{-2}`
        cosmo: astropy cosmology object
            Specifying a cosmology is required if `is_deltasigma` is True
        add: bool
            If `True`, adds the computed shears to the `galcat`

        Returns
        -------
        angsep: array_like
            Angular separation between lens and each source galaxy in radians
        tangential_component: array_like
            Tangential shear (or assimilated quantity) for each source galaxy
        cross_component: array_like
            Cross shear (or assimilated quantity) for each source galaxy
        """
        # Check is all the required data is available
        col_dict = {
            "ra_source": "ra",
            "dec_source": "dec",
            "shear1": shape_component1,
            "shear2": shape_component2,
        }
        if is_deltasigma:
            sigmac_colname = self.add_critical_surface_density(cosmo, use_pdz=use_pdz)
            col_dict.update({"sigma_c": sigmac_colname})
        cols = self._get_input_galdata(col_dict)

        # compute shears
        angsep, tangential_comp, cross_comp = compute_tangential_and_cross_components(
            is_deltasigma=is_deltasigma,
            ra_lens=self.ra,
            dec_lens=self.dec,
            geometry=geometry,
            validate_input=self.validate_input,
            coordinate_system=self.galcat.meta["coordinate_system"],
            **cols,
        )
        if add:
            self.galcat["theta"] = angsep
            self.galcat[tan_component] = tangential_comp
            self.galcat[cross_component] = cross_comp
            if is_deltasigma:
                sigmac_type = "effective" if use_pdz else "standard"
                self.galcat.meta[f"{tan_component}_sigmac_type"] = sigmac_type
                self.galcat.meta[f"{cross_component}_sigmac_type"] = sigmac_type
        return angsep, tangential_comp, cross_comp

    def compute_background_probability(
        self, use_pdz=False, add=True, p_background_name="p_background"
    ):
        r"""Probability for being a background galaxy

        Parameters
        ----------
        use_pdz : bool
            If True, computes the probability using the photoz pdf
        add : bool
            If True, add background probability columns to the galcat table
        p_background_name : str, optional
            User-defined name for the background probability column to be stored
            in the galcat table (i.e., if add=True)

        Returns
        -------
        p_background : array
            Probability for being a background galaxy
        """
        cols = self._get_input_galdata(
            {"pzpdf": "pzpdf", "pzbins": "pzbins"} if use_pdz else {"z_src": "z"}
        )
        p_background = compute_background_probability(
            self.z, use_pdz=use_pdz, validate_input=self.validate_input, **cols
        )
        if add:
            self.galcat[p_background_name] = p_background
        return p_background

    def compute_galaxy_weights(
        self,
        use_pdz=False,
        use_shape_noise=False,
        shape_component1="e1",
        shape_component2="e2",
        use_shape_error=False,
        shape_component1_err="e1_err",
        shape_component2_err="e2_err",
        weight_name="w_ls",
        cosmo=None,
        is_deltasigma=False,
        add=True,
    ):
        r"""Computes the individual lens-source pair weights

        Parameters
        ----------
        use_pdz : bool
            True for computing photometric weights
        use_shape_noise : bool
            True for considering shapenoise in the weight computation
        shape_component1: string
            column name : The measured shear (or reduced shear or ellipticity)
            of the source galaxies
        shape_component2: array
            column name : The measured shear (or reduced shear or ellipticity)
            of the source galaxies
        use_shape_error : bool
            True for considering measured shape error in the weight computation
        shape_component1_err: array
            column name : The measurement error on the 1st-component of ellipticity
            of the source galaxies
        shape_component2_err: array
            column name : The measurement error on the 2nd-component of ellipticity
            of the source galaxies
        weight_name : string
            Name of the new column for the weak lensing weights in the galcat table
        cosmo: clmm.Comology object, None
            CLMM Cosmology object.
        is_deltasigma: bool
            Indicates whether it is the excess surface density or the tangential shear
        add : bool
            If True, add weight column to the galcat table

        Returns
        -------
        w_ls: array
            the individual lens source pair weights
        """
        # input cols
        col_dict = {}
        if is_deltasigma:
            sigmac_colname = self.add_critical_surface_density(cosmo, use_pdz=use_pdz)
            col_dict.update({"sigma_c": sigmac_colname})
        if use_shape_noise:
            col_dict.update(
                {
                    "shape_component1": shape_component1,
                    "shape_component2": shape_component2,
                }
            )
        if use_shape_error:
            col_dict.update(
                {
                    "shape_component1_err": shape_component1_err,
                    "shape_component2_err": shape_component2_err,
                }
            )
        cols = self._get_input_galdata(col_dict)

        # computes weights
        w_ls = compute_galaxy_weights(
            is_deltasigma=is_deltasigma,
            use_shape_noise=use_shape_noise,
            use_shape_error=use_shape_error,
            validate_input=self.validate_input,
            **cols,
        )
        if add:
            self.galcat[weight_name] = w_ls
            if is_deltasigma:
                self.galcat.meta[f"{weight_name}_sigmac_type"] = (
                    "effective" if use_pdz else "standard"
                )
        return w_ls

    def draw_gal_z_from_pdz(self, zcol_out="z", overwrite=False, nobj=1, xmin=None, xmax=None):
        """Draw random redshifts from the photoz pdf for each galaxy
        of the galcat table.

        Parameters
        ----------
        zcol_out : string
            Name of the column of the galcat table where the random
            redshifts are to be stored. Default='z'
        overwrite : bool
            If True and if zcol_out already exists in the table, the column
            will be overwritten by the new random values
        nobj : int, optional
            Number of random samples to generate. Default is 1.
        xmin : float
            Lower bound to draw redshift. Default is the min(x_tab)
        xmax : float
            Upper bound to draw redshift. Default is the max(x_tab)

        Returns
        -------
        samples : ndarray
            Random points following the pdf_tab distribution
        """

        if zcol_out in self.galcat.columns and overwrite is False:
            raise TypeError(
                f"Column {zcol_out} already exists in galcat. \
                            Set overwrite=True to overwrite or use other column name"
            )

        zdata = self._get_input_galdata({"pzpdf": "pzpdf", "pzbins": "pzbins"})
        if self.galcat.pzpdf_info["type"] == "shared_bins":
            res = [
                _draw_random_points_from_tab_distribution(
                    zdata["pzbins"], pzpdf, nobj=nobj, xmin=xmin, xmax=xmax
                )
                for pzpdf in zdata["pzpdf"]
            ]
        else:
            res = [
                _draw_random_points_from_tab_distribution(
                    pzbin, pzpdf, nobj=nobj, xmin=xmin, xmax=xmax
                )
                for pzbin, pzpdf in zip(zdata["pzbins"], zdata["pzpdf"])
            ]

        self.galcat[zcol_out] = res
        return res

    def make_radial_profile(
        self,
        bin_units,
        bins=10,
        error_model="ste",
        cosmo=None,
        tan_component_in="et",
        cross_component_in="ex",
        tan_component_out="gt",
        cross_component_out="gx",
        tan_component_in_err=None,
        cross_component_in_err=None,
        include_empty_bins=False,
        gal_ids_in_bins=False,
        add=True,
        table_name="profile",
        overwrite=True,
        use_weights=False,
        weights_in="w_ls",
        weights_out="W_l",
    ):
        r"""Compute the shear or ellipticity profile of the cluster

        We assume that the cluster object contains information on the cross and
        tangential shears or ellipticities and angular separation of the source galaxies

        Calls `clmm.dataops.make_radial_profile` with the following arguments:
        components: `galcat` components (tan_component_in, cross_component_in, z)
        angsep: `galcat` theta
        angsep_units: 'radians'
        bin_units: `input` bin_units
        bins: `input` bins
        include_empty_bins: `input` include_empty_bins
        cosmo: `input` cosmo
        z_lens: cluster z

        Parameters
        ----------
        bin_units : str
            Units to use for the radial bins of the shear profile
            Allowed Options = ["radians", "deg", "arcmin", "arcsec", "kpc", "Mpc"]
            (letter case independent)
        bins : array_like, optional
            User defined bins to use for the shear profile. If a list is provided, use that as
            the bin edges. If a scalar is provided, create that many equally spaced bins between
            the minimum and maximum angular separations in bin_units. If nothing is provided,
            default to 10 equally spaced bins.
        error_model : str, optional
            Statistical error model to use for y uncertainties. (letter case independent)

                * `ste` - Standard error [=std/sqrt(n) in unweighted computation] (Default).
                * `std` - Standard deviation.

        cosmo: clmm.Comology object, None
            CLMM Cosmology object, used to convert angular separations to physical distances
        tan_component_in: string, optional
            Name of the tangential component column in `galcat` to be binned.
            Default: 'et'
        cross_component_in: string, optional
            Name of the cross component column in `galcat` to be binned.
            Default: 'ex'
        tan_component_out: string, optional
            Name of the tangetial component binned column to be added in profile table.
            Default: 'gt'
        cross_component_out: string, optional
            Name of the cross component binned profile column to be added in profile table.
            Default: 'gx'
        tan_component_in_err: string, None, optional
            Name of the tangential component error column in `galcat` to be binned.
            Default: None
        cross_component_in_err: string, None, optional
            Name of the cross component error column in `galcat` to be binned.
            Default: None
        include_empty_bins: bool, optional
            Also include empty bins in the returned table
        gal_ids_in_bins: bool, optional
            Also include the list of galaxies ID belonging to each bin in the returned table
        add: bool, optional
            Attach the profile to the cluster object
        table_name: str, optional
            Name of the profile table to be add as `cluster.table_name`.
            Default 'profile'
        overwrite: bool, optional
            Overwrite profile table.
            Default True
        use_weights: bool, optional
            If True, use the column `weights_in` in `galcat` as the weights
            Default: False
        weights_in: str, optional
            Name of the weights column in `galcat`
            Default: 'w_ls'

        Returns
        -------
        profile : GCData
            Output table containing the radius grid points, the tangential and cross shear
            profiles on that grid, and the errors in the two shear profiles. The errors are
            defined as the standard errors in each bin.
        """
        # Too many local variables (19/15)
        # pylint: disable=R0914

        if not all(
            t_ in self.galcat.columns for t_ in (tan_component_in, cross_component_in, "theta")
        ):
            raise TypeError(
                "Shear or ellipticity information is missing. Galaxy catalog must have tangential"
                "and cross shears (gt, gx) or ellipticities (et, ex). "
                "Run compute_tangential_and_cross_components first."
            )
        if "z" not in self.galcat.columns:
            raise TypeError("Missing galaxy redshifts!")
        # Compute the binned averages and associated errors
        profile_table, binnumber = make_radial_profile(
            [self.galcat[n].data for n in (tan_component_in, cross_component_in, "z")],
            angsep=self.galcat["theta"],
            angsep_units="radians",
            bin_units=bin_units,
            bins=bins,
            error_model=error_model,
            include_empty_bins=include_empty_bins,
            return_binnumber=True,
            cosmo=cosmo,
            z_lens=self.z,
            validate_input=self.validate_input,
            components_error=[
                None if n is None else self.galcat[n].data
                for n in (tan_component_in_err, cross_component_in_err, None)
            ],
            weights=self.galcat[weights_in].data if use_weights else None,
            coordinate_system=self.galcat.meta["coordinate_system"],
        )
        # Reaname table columns
        for i, name in enumerate([tan_component_out, cross_component_out, "z"]):
            profile_table.rename_column(f"p_{i}", name)
            profile_table.rename_column(f"p_{i}_err", f"{name}_err")
        # Reaname weights columns
        profile_table.rename_column("weights_sum", weights_out)
        # add galaxy IDs
        if gal_ids_in_bins:
            if "id" not in self.galcat.columns:
                raise TypeError("Missing galaxy IDs!")
            nbins = len(bins) - 1 if hasattr(bins, "__len__") else bins
            gal_ids = [list(self.galcat["id"][binnumber == i + 1]) for i in range(nbins)]
            if not include_empty_bins:
                gal_ids = [g_id for g_id in gal_ids if len(g_id) > 1]
            profile_table["gal_id"] = gal_ids
        if add:
            profile_table.update_cosmo_ext_valid(self.galcat, cosmo, overwrite=False)
            if hasattr(self, table_name):
                if overwrite:
                    warnings.warn(f"overwriting {table_name} table.")
                    delattr(self, table_name)
                else:
                    raise AttributeError(
                        f"table {table_name} already exists, "
                        "set overwrite=True or use another name."
                    )
            setattr(self, table_name, profile_table)
        return profile_table

    def plot_profiles(
        self,
        tangential_component="gt",
        tangential_component_error="gt_err",
        cross_component="gx",
        cross_component_error="gx_err",
        table_name="profile",
        xscale="linear",
        yscale="linear",
    ):
        """Plot shear profiles using `plotting.plot_profiles` function

        Parameters
        ----------
        tangential_component: str, optional
            Name of the column in the galcat Table corresponding to the tangential component of
            the shear or reduced shear (Delta Sigma not yet implemented). Default: 'gt'
        tangential_component_error: str, optional
            Name of the column in the galcat Table corresponding to the uncertainty in tangential
            component of the shear or reduced shear. Default: 'gt_err'
        cross_component: str, optional
            Name of the column in the galcat Table corresponding to the cross component of the
            shear or reduced shear. Default: 'gx'
        cross_component_error: str, optional
            Name of the column in the galcat Table corresponding to the uncertainty in the cross
            component of the shear or reduced shear. Default: 'gx_err'
        table_name: str, optional
            Name of the GalaxyCluster() `.profile` attribute. Default: 'profile'
        xscale:
            matplotlib.pyplot.xscale parameter to set x-axis scale (e.g. to logarithmic axis)
        yscale:
            matplotlib.pyplot.yscale parameter to set y-axis scale (e.g. to logarithmic axis)

        Returns
        -------
        fig:
            The matplotlib figure object that has been plotted to.
        axes:
            The matplotlib axes object that has been plotted to.
        """
        if not hasattr(self, table_name):
            raise ValueError(f"GalaxyClusters does not have a '{table_name}' table.")
        profile = getattr(self, table_name)
        for col in (tangential_component, cross_component):
            if col not in profile.columns:
                raise ValueError(f"Column for plotting '{col}' does not exist.")
        for col in (tangential_component_error, cross_component_error):
            if col not in profile.columns:
                warnings.warn(f"Column for plotting '{col}' does not exist.")
        return plot_profiles(
            rbins=profile["radius"],
            r_units=profile.meta["bin_units"],
            tangential_component=profile[tangential_component],
            tangential_component_error=(
                profile[tangential_component_error]
                if tangential_component_error in profile.columns
                else None
            ),
            cross_component=profile[cross_component],
            cross_component_error=(
                profile[cross_component_error] if cross_component_error in profile.columns else None
            ),
            xscale=xscale,
            yscale=yscale,
            tangential_component_label=tangential_component,
            cross_component_label=cross_component,
        )

    def set_ra_lower(self, ra_low=0):
        """
        Set window of values for cluster and galcat RA to [ra_low, ra_low+360[

        Parameters
        ----------
        ra_low: float
            Lower value for RA range, can be -180 or 0

        """
        if ra_low not in (-180.0, 0.0):
            raise ValueError("ra_low must be -180 or 0")
        self.ra += 360.0 if self.ra < ra_low else 0
        self.ra -= 360.0 if self.ra >= ra_low + 360.0 else 0
        if "ra" in self.galcat.columns:
            self.galcat["ra"][self.galcat["ra"] < ra_low] += 360.0
            self.galcat["ra"][self.galcat["ra"] >= ra_low + 360.0] -= 360.0

    def update_coordinate_system(self, coordinate_system, ellipticity_columns=()):
        """Updates coordinate_system metadata of the galcat and converts ellipticity

        Parameters
        ----------
        coordinate_system : str
            Coordinate system of the ellipticity components. Must be either 'celestial' or
            euclidean'. See https://doi.org/10.48550/arXiv.1407.7676 section 5.1 for more details.
        ellipticity_columns : tuple
            Components to be converted, must be in data.

        Returns
        -------
        None
        """
        self.galcat.update_coordinate_system(coordinate_system, ellipticity_columns)



"""@file miscentering.py
Model functions with miscentering
"""

# Functions to model halo profiles

import numpy as np
from scipy.integrate import dblquad, quad, tplquad


def integrand_surface_density_nfw(theta, r_proj, r_mis, r_s):
    r"""Computes integrand for surface mass density with the NFW profile.

    Parameters
    ----------
    theta : float
        Angle of polar coordinates of the miscentering direction.
    r_proj : float
        Projected radial position from the cluster center in :math:`M\!pc`.
    r_mis : float
        Projected miscenter distance in :math:`M\!pc`.
    r_s : float
        Scale radius

    Returns
    -------
    float
        2D projected density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    r_norm = np.sqrt(r_proj**2.0 + r_mis**2.0 - 2.0 * r_proj * r_mis * np.cos(theta)) / r_s

    r2m1 = r_norm**2.0 - 1.0
    if r_norm < 1:
        sqrt_r2m1 = np.sqrt(-r2m1)
        res = np.arcsinh(sqrt_r2m1 / r_norm) / (-r2m1) ** (3.0 / 2.0) + 1.0 / r2m1
    elif r_norm > 1:
        sqrt_r2m1 = np.sqrt(r2m1)
        res = -np.arcsin(sqrt_r2m1 / r_norm) / (r2m1) ** (3.0 / 2.0) + 1.0 / r2m1
    else:
        res = 1.0 / 3.0
    return res


def integrand_surface_density_einasto(r_par, theta, r_proj, r_mis, r_s, alpha_ein):
    r"""Computes integrand for surface mass density with the Einasto profile.

    Parameters
    ----------
    r_par : float
        Parallel radial position from the cluster center in :math:`M\!pc`.
    theta : float
        Angle of polar coordinates of the miscentering direction.
    r_proj : array_like
        Projected radial position from the cluster center in :math:`M\!pc`.
    r_mis : float
        Projected miscenter distance in :math:`M\!pc`.
    r_s : float
        Scale radius
    alpha_ein : float
        Einasto slope

    Returns
    -------
    float
        2D projected density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    # Projected surface mass density element for numerical integration
    r_norm = (
        np.sqrt(r_par**2.0 + r_proj**2.0 + r_mis**2.0 - 2.0 * r_proj * r_mis * np.cos(theta)) / r_s
    )

    return np.exp(-2.0 * (r_norm**alpha_ein - 1.0) / alpha_ein)


def integrand_surface_density_hernquist(theta, r_proj, r_mis, r_s):
    r"""Computes integrand for surface mass density with the Hernquist profile.

    Parameters
    ----------
    theta : float
        Angle of polar coordinates of the miscentering direction.
    r_proj : float
        Projected radial position from the cluster center in :math:`M\!pc`.
    r_mis : float
        Projected miscenter distance in :math:`M\!pc`.
    r_s : float
        Scale radius

    Returns
    -------
    float
        2D projected density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    r_norm = np.sqrt(r_proj**2.0 + r_mis**2.0 - 2.0 * r_proj * r_mis * np.cos(theta)) / r_s

    r2m1 = r_norm**2.0 - 1.0
    if r_norm < 1:
        res = -3 / r2m1**2 + (r2m1 + 3) * np.arcsinh(np.sqrt(-r2m1) / r_norm) / (-r2m1) ** 2.5
    elif r_norm > 1:
        res = -3 / r2m1**2 + (r2m1 + 3) * np.arcsin(np.sqrt(r2m1) / r_norm) / (r2m1) ** 2.5
    else:
        res = 4.0 / 15.0
    return res


def integrand_mean_surface_density_nfw(theta, r_proj, r_mis, r_s):
    r"""Computes integrand for mean surface mass density with the NFW profile.

    Parameters
    ----------
    theta : float
        Angle of polar coordinates of the miscentering direction.
    r_proj : float
        Projected radial position from the cluster center in :math:`M\!pc`.
    r_mis : float
        Projected miscenter distance in :math:`M\!pc`.
    r_s : float
        Scale radius

    Returns
    -------
    float
        Mean surface density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    return r_proj * integrand_surface_density_nfw(theta, r_proj, r_mis, r_s)


def integrand_mean_surface_density_einasto(r_par, theta, r_proj, r_mis, r_s, alpha_ein):
    r"""Computes integrand for mean surface mass density with the Einasto profile.

    Parameters
    ----------
    r_par : float
        Parallel radial position from the cluster center in :math:`M\!pc`.
    theta : float
        Angle of polar coordinates of the miscentering direction.
    r_proj : float
        Projected radial position from the cluster center in :math:`M\!pc`.
    r_mis : float
        Projected miscenter distance in :math:`M\!pc`.
    r_s : float
        Scale radius
    alpha_ein : float
        Einasto slope

    Returns
    -------
    float
        Mean surface density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    return r_proj * integrand_surface_density_einasto(r_par, theta, r_proj, r_mis, r_s, alpha_ein)


def integrand_mean_surface_density_hernquist(theta, r_proj, r_mis, r_s):
    r"""Computes integrand for mean surface mass density with the Hernquist profile.

    Parameters
    ----------
    theta : float
        Angle of polar coordinates of the miscentering direction.
    r_proj : float
        Projected radial position from the cluster center in :math:`M\!pc`.
    r_mis : float
        Projected miscenter distance in :math:`M\!pc`.
    r_s : float
        Scale radius

    Returns
    -------
    float
        Mean surface density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    return r_proj * integrand_surface_density_hernquist(theta, r_proj, r_mis, r_s)


def integrate_azimuthially_miscentered_surface_density(
    r_proj, r_mis, integrand, norm, aux_args, extra_integral
):
    r"""Integrates azimuthally the miscentered surface mass density kernel.

    Parameters
    ----------
    r_proj : float, array_like
        Projected radial position from the cluster center in :math:`M\!pc`.
    r_mis : float
        Projected miscenter distance in :math:`M\!pc`.
    integrand : function
        Function to be integrated
    norm : float
        Normalization value for integral
    aux_args : list
        Auxiliary arguments used in the integral
    extra_integral : bool
        Additional dimension for the integral

    Returns
    -------
    float, numpy.ndarray
        2D projected density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    args_list = [(r, r_mis, *aux_args) for r in np.atleast_1d(r_proj)]
    if extra_integral:
        res = [
            dblquad(integrand, 0.0, np.pi, 0, np.inf, args=args, epsrel=1e-6)[0]
            for args in args_list
        ]
    else:
        res = [quad(integrand, 0.0, np.pi, args=args, epsrel=1e-6)[0] for args in args_list]

    res = np.array(res) * norm / np.pi

    if not np.iterable(r_proj):
        return res[0]
    return res


def integrate_azimuthially_miscentered_mean_surface_density(
    r_proj, r_mis, integrand, norm, aux_args, extra_integral
):
    r"""Integrates azimuthally the miscentered mean surface mass density kernel.

    Parameters
    ----------
    r_proj : float, array_like
        Projected radial position from the cluster center in :math:`M\!pc`.
    r_mis : float
        Projected miscenter distance in :math:`M\!pc`.
    integrand : function
        Function to be integrated
    norm : float
        Normalization value for integral
    aux_args : list
        Auxiliary arguments used in the integral
    extra_integral : bool
        Additional dimension for the integral

    Returns
    -------
    float, numpy.ndarray
        Mean surface density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    r_proj_use = np.atleast_1d(r_proj)
    r_lower = np.zeros_like(r_proj_use)
    r_lower[1:] = r_proj_use[:-1]
    args = (r_mis, *aux_args)

    if extra_integral:
        res = [
            tplquad(integrand, r_low, r_high, 0, np.pi, 0, np.inf, args=args, epsrel=1e-6)[0]
            for r_low, r_high in zip(r_lower, r_proj_use)
        ]
    else:
        res = [
            dblquad(integrand, r_low, r_high, 0, np.pi, args=args, epsrel=1e-6)[0]
            for r_low, r_high in zip(r_lower, r_proj_use)
        ]

    if not np.iterable(r_proj):
        return res[0] * norm * 2 / np.pi / r_proj**2
    return np.cumsum(res) * norm * 2 / np.pi / r_proj**2



"""Provide a consistent set of constants to use through CLMM"""

from enum import Enum


class Constants(Enum):
    """A set of constants for consistency throughout the
    code and dependencies."""

    CLIGHT = 299792458.0
    """ Speed of light (m/s)

    Source: CODATA 2018
    """

    CLIGHT_KMS = CLIGHT * 1.0e-3
    """ Speed of light (km/s)

    Source: CODATA 2018
    """

    GNEWT = 6.67430e-11
    """ Newton's constant (m^3/kg/s^2)

    Source: CODATA 2018
    """

    PC_TO_METER = 3.085677581491367e16
    """ parsec to meter (m)

    Source: IAU 2015
    """

    GNEWT_SOLAR_MASS = 1.3271244e20
    """ G x Solar mass (m^3/s^2)

    Source: IAU 2015
    """

    SOLAR_MASS = GNEWT_SOLAR_MASS / GNEWT
    """ Solar mass (kg)

    Source: IAU 2015/CODATA 2018
    """



"""@file generic.py
Model independent theory functions
"""

# Functions to model halo profiles

import warnings

import numpy as np
from scipy.optimize import fsolve
from scipy.special import gamma, gammainc

# functions that are general to all backends


def compute_reduced_shear_from_convergence(shear, convergence):
    """Calculates reduced shear from shear and convergence

    Parameters
    ----------
    shear : array_like, float
        Shear
    convergence : array_like, float
        Convergence

    Returns
    -------
    g : array_like, float
        Reduced shear
    """
    reduced_shear = np.array(shear) / (1.0 - np.array(convergence))
    return reduced_shear


def compute_magnification_bias_from_magnification(magnification, alpha):
    r"""Computes magnification bias from magnification :math:`\mu` and slope parameter
    :math:`\alpha` as :

    .. math::
        \mu^{\alpha - 1}

    The alpha parameter depends on the source sample and is computed as the slope of the
    cummulative numer counts at a given magnitude:

    .. math::
        \alpha \equiv \alpha(f) = - \frac{\mathrm d}{\mathrm d \log{f}} \log{n_0(>f)}

    or,

    .. math::
        \alpha \equiv \alpha(m) = 2.5 \frac{\mathrm d}{\mathrm d m} \log{n_0(<m)}

    see e.g.  Bartelmann & Schneider 2001; Umetsu 2020

    Parameters
    ----------
    magnification : array_like
        Magnification
    alpha : array like
        Source cummulative number density slope

    Returns
    -------
    magnification bias : array_like
        magnification bias
    """
    if np.any(np.array(magnification) < 0):
        warnings.warn(
            "Magnification is negative for certain radii, \
                    returning nan for magnification bias in this case."
        )
    return np.array(magnification) ** (np.array([alpha]).T - 1)


def compute_rdelta(mdelta, redshift, cosmo, massdef="mean", delta_mdef=200):
    r"""Computes the radius for mdelta

    .. math::
        r_\Delta=\left(\frac{3 M_\Delta}{4 \pi \Delta \rho_{bkg}(z)}\right)^{1/3}

    Parameters
    ----------
    mdelta : float
        Mass in units of :math:`M_\odot`
    redshift : float
        Redshift of the cluster
    cosmo : clmm.Cosmology
        Cosmology object
    massdef : str, None
        Profile mass definition ("mean", "critical", "virial").
    delta_mdef : int, None
        Mass overdensity definition.

    Returns
    -------
    float
        Radius in :math:`M\!pc`.
    """
    # Check massdef
    if massdef == "mean":
        rho = cosmo.get_rho_m
    elif massdef in ("critical", "virial"):
        rho = cosmo.get_rho_c
    else:
        raise ValueError(f"massdef(={massdef}) must be mean, critical or virial")
    return ((3.0 * mdelta) / (4.0 * np.pi * delta_mdef * rho(redshift))) ** (1.0 / 3.0)


def compute_profile_mass_in_radius(
    r3d,
    redshift,
    cosmo,
    mdelta,
    cdelta,
    massdef="mean",
    delta_mdef=200,
    halo_profile_model="nfw",
    alpha=None,
):
    r"""Computes the mass inside a given radius of the profile.
    The mass is calculated as

    .. math::
        M(<\text{r3d}) = M_{\Delta}\;
        \frac{f\left(\frac{\text{r3d}}{r_{\Delta}/c_{\Delta}}\right)}{f(c_{\Delta})},

    where :math:`f(x)` for the different models are

    NFW:

    .. math::
        \quad \ln(1+x)-\frac{x}{1+x}

    Einasto: (:math:`\gamma` is the lower incomplete gamma function)

    .. math::
        \gamma(\frac{3}{\alpha}, \frac{2}{\alpha}x^{\alpha})

    Hernquist:

    .. math::
        \left(\frac{x}{1+x}\right)^2


    Parameters
    ----------
    r3d : array_like, float
        Radial position from the cluster center in :math:`M\!pc`.
    refshift : float
        Redshift of the cluster
    mdelta : float
        Mass of the profile in units of :math:`M_\odot`
    cdelta : float
        Concentration of the profile.
    massdef : str, None
        Profile mass definition ("mean", "critical", "virial").
    delta_mdef : int, None
        Mass overdensity definition.
    halo_profile_model : str
        Profile model parameterization ("nfw", "einasto", "hernquist").
    alpha : float, None
        Einasto slope, required when `halo_profile_model='einasto'`.

    Returns
    -------
    array_like, float
        Mass in units of :math:`M_\odot`
    """
    # pylint: disable=unnecessary-lambda-assignment
    rdelta = compute_rdelta(mdelta, redshift, cosmo, massdef, delta_mdef)
    # Check halo_profile_model
    if halo_profile_model == "nfw":
        prof_integ = lambda c: np.log(1.0 + c) - c / (1.0 + c)
    elif halo_profile_model == "einasto":
        if alpha is None:
            raise ValueError("alpha must be provided when Einasto profile is selected!")
        prof_integ = lambda c: gamma(3.0 / alpha) * gammainc(3.0 / alpha, 2.0 / alpha * c**alpha)
    elif halo_profile_model == "hernquist":
        prof_integ = lambda c: (c / (1.0 + c)) ** 2.0
    else:
        raise ValueError(
            f"halo_profile_model=(={halo_profile_model=}) must be " "nfw, einasto, or hernquist!"
        )
    r3d_norm = np.array(r3d) / (rdelta / cdelta)
    return mdelta * prof_integ(r3d_norm) / prof_integ(cdelta)


def convert_profile_mass_concentration(
    mdelta,
    cdelta,
    redshift,
    cosmo,
    massdef,
    delta_mdef,
    halo_profile_model,
    massdef2=None,
    delta_mdef2=None,
    halo_profile_model2=None,
    alpha=None,
    alpha2=None,
):
    r"""
    Parameters
    ----------
    mdelta : float
        Mass of the profile in units of :math:`M_\odot`
    cdelta : float
        Concentration of the profile.
    refshift : float
        Redshift of the cluster
    cosmo : clmm.Cosmology
        Cosmology object
    massdef : str, None
        Input profile mass definition ("mean", "critical", "virial").
    delta_mdef : int, None
        Input mass overdensity definition.
    halo_profile_model : str, None
        Input profile model parameterization ("nfw", "einasto", "hernquist").
    massdef2 : str, None
        Profile mass definition to convert to ("mean", "critical", "virial").
        If None, `massdef2=massdef`.
    delta_mdef2 : int, None
        Mass overdensity definition to convert to.
        If None, `delta_mdef2=delta_mdef`.
    halo_profile_model2 : str, None
        Profile model parameterization to convert to ("nfw", "einasto", "hernquist").
        If None, `halo_profile_model2=halo_profile_model`.
    alpha : float, None
        Input Einasto slope when `halo_profile_model='einasto'`.
    alpha2 : float, None
        Einasto slope to convert to when `halo_profile_model='einasto'`.
        If None, `alpha2=alpha`.

    Returns
    -------
    HaloProfile:
        HaloProfile object
    """
    # pylint: disable=unused-argument
    rdelta = compute_rdelta(mdelta, redshift, cosmo, massdef, delta_mdef)
    # Prep other args
    loc, keys = locals(), ("massdef", "delta_mdef", "halo_profile_model", "alpha")
    kwargs = {key: loc[key] for key in keys}
    kwargs2 = {key: (loc[key] if loc[f"{key}2"] is None else loc[f"{key}2"]) for key in keys}

    # Eq. to solve
    def delta_mass(params):
        mdelta2, cdelta2 = params
        rdelta2 = compute_rdelta(
            mdelta2, redshift, cosmo, kwargs2["massdef"], kwargs2["delta_mdef"]
        )
        mdelta2_rad1 = compute_profile_mass_in_radius(
            rdelta, redshift, cosmo, mdelta2, cdelta2, **kwargs2
        )
        mdelta1_rad2 = compute_profile_mass_in_radius(
            rdelta2, redshift, cosmo, mdelta, cdelta, **kwargs
        )
        return mdelta - mdelta2_rad1, mdelta2 - mdelta1_rad2

    # Iterate 2 times:
    mdelta2, cdelta2 = fsolve(func=delta_mass, x0=[mdelta, cdelta])
    mdelta2, cdelta2 = fsolve(func=delta_mass, x0=[mdelta2, cdelta2])

    return mdelta2, cdelta2



"""Data operation for polar/azimuthal averages in radial bins and weights"""

import warnings

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from ..gcdata import GCData
from ..utils import (
    _integ_pzfuncs,
    _validate_coordinate_system,
    _validate_dec,
    _validate_is_deltasigma_sigma_c,
    _validate_ra,
    arguments_consistency,
    compute_radial_averages,
    convert_units,
    make_bins,
    validate_argument,
)


def compute_tangential_and_cross_components(
    ra_lens,
    dec_lens,
    ra_source,
    dec_source,
    shear1,
    shear2,
    coordinate_system=None,
    geometry="curve",
    is_deltasigma=False,
    sigma_c=None,
    validate_input=True,
):
    r"""Computes tangential- and cross- components for shear or ellipticity

    To do so, we need the right ascension and declination of the lens and of all of the sources.
    We also need the two shape components of all of the sources.

    These quantities can be handed to `compute_tangential_and_cross_components` in two ways

    1. Pass in each as parameters::

        compute_tangential_and_cross_components(ra_lens, dec_lens, ra_source, dec_source,
        shape_component1, shape_component2)

    2. As a method of `GalaxyCluster`::

        cluster.compute_tangential_and_cross_components()

    The angular separation between the source and the lens, :math:`\theta`, and the azimuthal
    position of the source relative to the lens, :math:`\phi`, are computed within the function
    and the angular separation is returned.

    In the flat sky approximation, these angles are calculated using (_lens: lens,
    _source: source, RA is from right to left)

    .. math::
        \theta^2 = & \left(\delta_s-\delta_l\right)^2+
        \left(\alpha_l-\alpha_s\right)^2\cos^2(\delta_l)\\
        \tan\phi = & \frac{\delta_s-\delta_l}{\left(\alpha_l-\alpha_s\right)\cos(\delta_l)}

    The tangential, :math:`g_t`, and cross, :math:`g_x`, ellipticity/shear components are
    calculated using the two ellipticity/shear components :math:`g_1` and :math:`g_2` of the
    source galaxies, following Eq.7 and Eq.8 in Schrabback et al. (2018), arXiv:1611:03866 which
    is consistent with arXiv:0509252

    .. math::
        g_t =& -\left( g_1\cos\left(2\phi\right)+g_2\sin\left(2\phi\right)\right)\\
        g_x =& g_1 \sin\left(2\phi\right)-g_2\cos\left(2\phi\right)

    Finally, if  the critical surface density (:math:`\Sigma_\text{crit}`) is provided, an estimate
    of the excess surface density :math:`\widehat{\Delta\Sigma}` is obtained from

    .. math::
        \widehat{\Delta\Sigma_{t,x}} = g_{t,x} \times \Sigma_\text{crit}(z_l, z_{\text{src}})

    If :math:`g_{t,x}` correspond to the shear, the above expression is an accurate. However, if
    :math:`g_{t,x}` correspond to ellipticities or reduced shear, this expression only gives an
    estimate :math:`\widehat{\Delta\Sigma_{t,x}}`, valid only in the weak lensing regime.

    Parameters
    ----------
    ra_lens: float
        Right ascension of the lensing cluster in degrees
    dec_lens: float
        Declination of the lensing cluster in degrees
    ra_source: array
        Right ascensions of each source galaxy in degrees
    dec_source: array
        Declinations of each source galaxy in degrees
    shear1: array
        The measured shear (or reduced shear or ellipticity) of the source galaxies
    shear2: array
        The measured shear (or reduced shear or ellipticity) of the source galaxies
    coordinate_system: str, optional
        Coordinate system of the ellipticity components. Must be either 'celestial' or
        euclidean'. See https://doi.org/10.48550/arXiv.1407.7676 section 5.1 for more details.
        If not set, defaults to 'euclidean'.
    geometry: str, optional
        Sky geometry to compute angular separation.
        Options are curve (uses astropy) or flat.
    is_deltasigma: bool
        If `True`, the tangential and cross components returned are multiplied by Sigma_crit.
        It requires `sigma_c` argument. Results in units of :math:`M_\odot\ Mpc^{-2}`
    sigma_c : None, array_like
        Critical (effective) surface density in units of :math:`M_\odot\ Mpc^{-2}`.
        Used only when is_deltasigma=True.
    validate_input: bool
        Validade each input argument

    Returns
    -------
    angsep: array_like
        Angular separation between lens and each source galaxy in radians
    tangential_component: array_like
        Tangential shear (or assimilated quantity) for each source galaxy
    cross_component: array_like
        Cross shear (or assimilated quantity) for each source galaxy
    """
    # pylint: disable-msg=too-many-locals
    # Note: we make these quantities to be np.array so that a name is not passed from astropy
    # columns
    if coordinate_system is None:
        warnings.warn("coordinate_system not set, defaulting to 'euclidean'")
        coordinate_system = "euclidean"

    if validate_input:
        _validate_ra(locals(), "ra_source", True)
        _validate_dec(locals(), "dec_source", True)
        _validate_ra(locals(), "ra_lens", True)
        _validate_dec(locals(), "dec_lens", True)
        validate_argument(locals(), "shear1", "float_array")
        validate_argument(locals(), "shear2", "float_array")
        validate_argument(locals(), "geometry", str)
        validate_argument(locals(), "sigma_c", "float_array", none_ok=True)
        _validate_coordinate_system(locals(), "coordinate_system")
        ra_source_, dec_source_, shear1_, shear2_ = arguments_consistency(
            [ra_source, dec_source, shear1, shear2],
            names=("Ra", "Dec", "Shear1", "Shear2"),
            prefix="Tangential- and Cross- shape components sources",
        )
        _validate_is_deltasigma_sigma_c(is_deltasigma, sigma_c)
    elif np.iterable(ra_source):
        ra_source_, dec_source_, shear1_, shear2_ = (
            np.array(col) for col in [ra_source, dec_source, shear1, shear2]
        )
    else:
        ra_source_, dec_source_, shear1_, shear2_ = (
            ra_source,
            dec_source,
            shear1,
            shear2,
        )
    # Compute the lensing angles
    if geometry == "flat":
        angsep, phi = _compute_lensing_angles_flatsky(
            ra_lens, dec_lens, ra_source_, dec_source_, coordinate_system=coordinate_system
        )
    elif geometry == "curve":
        angsep, phi = _compute_lensing_angles_astropy(
            ra_lens, dec_lens, ra_source_, dec_source_, coordinate_system=coordinate_system
        )
    else:
        raise NotImplementedError(f"Sky geometry {geometry} is not currently supported")
    # Compute the tangential and cross shears
    tangential_comp = _compute_tangential_shear(shear1_, shear2_, phi)
    cross_comp = _compute_cross_shear(shear1_, shear2_, phi)

    if sigma_c is not None:
        _sigma_c_arr = np.array(sigma_c)
        tangential_comp *= _sigma_c_arr
        cross_comp *= _sigma_c_arr

    return angsep, tangential_comp, cross_comp


def compute_background_probability(
    z_lens, z_src=None, use_pdz=False, pzpdf=None, pzbins=None, validate_input=True
):
    r"""Probability for being a background galaxy, defined by:

        .. math::
            P(z_s > z_l) = \int_{z_l}^{+\infty} dz_s \; p_{\text{photoz}}(z_s),

    when the photometric probability density functions (:math:`p_{\text{photoz}}(z_s)`) are
    provided. In the case of true redshifts, it returns 1 if :math:`z_s > z_l` else returns 0.


    Parameters
    ----------
    z_lens: float
        Redshift of the lens.
    z_src: array, optional
        Redshift of the source. Used only if pzpdf=pzbins=None.

    Returns
    -------
    p_background : array
        Probability for being a background galaxy
    """
    if validate_input:
        validate_argument(locals(), "z_lens", float, argmin=0, eqmin=True)
        validate_argument(locals(), "z_src", "float_array", argmin=0, eqmin=True, none_ok=True)

    if use_pdz is False:
        if z_src is None:
            raise ValueError("z_src must be provided.")
        p_background = np.array(z_src > z_lens, dtype=float)
    else:
        if pzpdf is None or pzbins is None:
            raise ValueError("pzbins must be provided with pzpdf.")
        p_background = _integ_pzfuncs(pzpdf, pzbins, z_lens)

    return p_background


def compute_galaxy_weights(
    use_shape_noise=False,
    shape_component1=None,
    shape_component2=None,
    use_shape_error=False,
    shape_component1_err=None,
    shape_component2_err=None,
    is_deltasigma=False,
    sigma_c=None,
    validate_input=True,
):
    r"""Computes the individual lens-source pair weights

    The weights :math:`w_{ls}` express as : :math:`w_{ls} = w_{ls, \text{geo}} \times w_{ls,
    \text{shape}}`, following E. S. Sheldon et al. (2003), arXiv:astro-ph/0312036:

    1. If computed for shear, the geometrical weights :math:`w_{ls, \text{geo}}` are equal to 1. If
    computed for :math:`\Delta \Sigma`, it depends on lens and source redshift information via the
    critical surface density. This component can be expressed as:

        .. math::
            w_{ls, \text{geo}} = \Sigma_\text{crit}(z_l, z_{\text{src}})^{-2}\;.

        when only redshift point estimates are provided, or as:


        .. math::
            w_{ls, \text{geo}} = \Sigma_\text{crit}^\text{eff}(z_l, z_{\text{src}})^{-2}
            = \left[\int_{\delta + z_l} dz_s \; p_{\text{photoz}}(z_s)
            \Sigma_\text{crit}(z_l, z_s)^{-1}\right]^2

        when the redshift pdf of each source, :math:`p_{\text{photoz}}(z_s)`, is known.

    2. The shape weight :math:`w_{ls,{\text{shape}}}` depends on shapenoise and/or shape
    measurement errors

        .. math::
            w_{ls, \text{shape}}^{-1} = \sigma_{\text{shapenoise}}^2 +
            \sigma_{\text{measurement}}^2


    Parameters
    ----------
    is_deltasigma: bool
        If `False`, weights are computed for shear, else weights are computed for
        :math:`\Delta \Sigma`.
    sigma_c : None, array_like
        Critical (effective) surface density in units of :math:`M_\odot\ Mpc^{-2}`.
        Used only when is_deltasigma=True.
    use_shape_noise: bool
        If `True` shape noise is included in the weight computation. It then requires
        `shape_componenet{1,2}` to be provided. Default: False.
    shape_component1: array_like
        The measured shear (or reduced shear or ellipticity) of the source galaxies,
        used if `use_shapenoise=True`
    shape_component2: array_like
        The measured shear (or reduced shear or ellipticity) of the source galaxies,
        used if `use_shapenoise=True`
    use_shape_error: bool
        If `True` shape errors are included in the weight computation. It then requires
        shape_component{1,2}_err` to be provided. Default: False.
    shape_component1_err: array_like
        The measurement error on the 1st-component of ellipticity of the source galaxies,
        used if `use_shape_error=True`
    shape_component2_err: array_like
        The measurement error on the 2nd-component of ellipticity of the source galaxies,
        used if `use_shape_error=True`
    validate_input: bool
        Validade each input argument

    Returns
    -------
    w_ls: array_like
        Individual lens source pair weights
    """
    if validate_input:
        validate_argument(locals(), "sigma_c", "float_array", none_ok=True)
        validate_argument(locals(), "shape_component1", "float_array", none_ok=True)
        validate_argument(locals(), "shape_component2", "float_array", none_ok=True)
        validate_argument(locals(), "shape_component1_err", "float_array", none_ok=True)
        validate_argument(locals(), "shape_component2_err", "float_array", none_ok=True)
        validate_argument(locals(), "use_shape_noise", bool)
        validate_argument(locals(), "use_shape_error", bool)
        arguments_consistency(
            [shape_component1, shape_component2],
            names=("shape_component1", "shape_component2"),
            prefix="Shape components sources",
        )
        _validate_is_deltasigma_sigma_c(is_deltasigma, sigma_c)

    # computing w_ls_geo
    w_ls_geo = 1.0
    if sigma_c is not None:
        w_ls_geo /= np.array(sigma_c) ** 2

    # computing w_ls_shape
    err_e2 = 0

    if use_shape_noise:
        if shape_component1 is None or shape_component2 is None:
            raise ValueError(
                "With the shape noise option, the source shapes "
                "`shape_component_{1,2}` must be specified"
            )
        err_e2 += np.std(shape_component1) ** 2 + np.std(shape_component2) ** 2
    if use_shape_error:
        if shape_component1_err is None or shape_component2_err is None:
            raise ValueError(
                "With the shape error option, the source shapes errors"
                "`shape_component_err{1,2}` must be specified"
            )
        err_e2 += shape_component1_err**2
        err_e2 += shape_component2_err**2

    if hasattr(err_e2, "__len__"):
        w_ls_shape = np.ones(len(err_e2))
        w_ls_shape[err_e2 > 0] = 1.0 / err_e2[err_e2 > 0]
    else:
        w_ls_shape = 1.0 / err_e2 if err_e2 > 0 else 1.0

    w_ls = w_ls_shape * w_ls_geo

    return w_ls


def _compute_lensing_angles_flatsky(
    ra_lens, dec_lens, ra_source_list, dec_source_list, coordinate_system="euclidean"
):
    r"""Compute the angular separation between the lens and the source and the azimuthal
    angle from the lens to the source in radians.

    In the flat sky approximation, these angles are calculated using

    .. math::
        \theta = \sqrt{\left(\delta_s-\delta_l\right)^2+
        \left(\alpha_l-\alpha_s\right)^2\cos^2(\delta_l)}

        \tan\phi = \frac{\delta_s-\delta_l}{\left(\alpha_l-\alpha_s\right)\cos(\delta_l)}

    For extended descriptions of parameters, see `compute_shear()` documentation.

    Parameters
    ----------
    ra_lens: float
        Right ascension of the lensing cluster in degrees
    dec_lens: float
        Declination of the lensing cluster in degrees
    ra_source_list: array
        Right ascensions of each source galaxy in degrees
    dec_source_list: array
        Declinations of each source galaxy in degrees
    coordinate_system: str, optional
        Coordinate system of the ellipticity components. Must be either 'celestial' or
        euclidean'. See https://doi.org/10.48550/arXiv.1407.7676 section 5.1 for more details.
        Default is 'euclidean'.

    Returns
    -------
    angsep: array
        Angular separation between the lens and the source in radians
    phi: array
        Azimuthal angle from the lens to the source in radians
    """
    delta_ra = np.radians(ra_source_list - ra_lens)
    # Put angles between -pi and pi
    delta_ra -= np.round(delta_ra / (2.0 * np.pi)) * 2.0 * np.pi
    deltax = delta_ra * np.cos(np.radians(dec_lens))
    deltay = np.radians(dec_source_list - dec_lens)
    # Ensure that abs(delta ra) < pi
    # deltax[deltax >= np.pi] = deltax[deltax >= np.pi]-2.*np.pi
    # deltax[deltax < -np.pi] = deltax[deltax < -np.pi]+2.*np.pi
    angsep = np.sqrt(deltax**2 + deltay**2)
    phi = np.arctan2(deltay, -deltax)
    # Forcing phi to be zero everytime angsep is zero. This is necessary due to arctan2 features.
    if np.iterable(phi):
        phi[angsep == 0.0] = 0.0
    else:
        phi = 0.0 if angsep == 0.0 else phi
    if coordinate_system == "celestial":
        phi = np.pi - phi
    if np.any(angsep > np.pi / 180.0):
        warnings.warn("Using the flat-sky approximation with separations >1 deg may be inaccurate")
    return angsep, phi


def _compute_lensing_angles_astropy(
    ra_lens, dec_lens, ra_source_list, dec_source_list, coordinate_system="euclidean"
):
    r"""Compute the angular separation between the lens and the source and the azimuthal
    angle from the lens to the source in radians.

    Parameters
    ----------
    ra_lens: float
        Right ascension of the lensing cluster in degrees
    dec_lens: float
        Declination of the lensing cluster in degrees
    ra_source_list: array
        Right ascensions of each source galaxy in degrees
    dec_source_list: array
        Declinations of each source galaxy in degrees
    coordinate_system: str, optional
        Coordinate system of the ellipticity components. Must be either 'celestial' or
        euclidean'. See https://doi.org/10.48550/arXiv.1407.7676 section 5.1 for more details.
        Default is 'euclidean'.

    Returns
    -------
    angsep: array
        Angular separation between the lens and the source in radians
    phi: array
        Azimuthal angle from the lens to the source in radians
    """
    sk_lens = SkyCoord(ra_lens * u.deg, dec_lens * u.deg, frame="icrs")
    sk_src = SkyCoord(ra_source_list * u.deg, dec_source_list * u.deg, frame="icrs")
    angsep, phi = sk_lens.separation(sk_src).rad, sk_lens.position_angle(sk_src).rad
    # Transformations for phi to have same orientation as _compute_lensing_angles_flatsky
    phi += 0.5 * np.pi
    if np.iterable(phi):
        phi[phi > np.pi] -= 2 * np.pi
        phi[angsep == 0] = 0
    else:
        phi -= 2 * np.pi if phi > np.pi else 0
        phi = 0 if angsep == 0 else phi
    if coordinate_system == "celestial":
        phi = np.pi - phi
    return angsep, phi


def _compute_tangential_shear(shear1, shear2, phi):
    r"""Compute the tangential shear given the two shears and azimuthal positions for
    a single source or list of sources.

    We compute the tangential shear following Eq. 7 of Schrabback et al. 2018, arXiv:1611:03866

    .. math::
        g_t = -\left( g_1\cos\left(2\phi\right)+g_2\sin\left(2\phi\right)\right)

    For extended descriptions of parameters, see `compute_shear()` documentation.
    """
    return -(shear1 * np.cos(2.0 * phi) + shear2 * np.sin(2.0 * phi))


def _compute_cross_shear(shear1, shear2, phi):
    r"""Compute the cross shear given the two shears and azimuthal position for a single
    source of list of sources.

    We compute the cross shear following Eq. 8 of Schrabback et al. 2018, arXiv:1611:03866
    also checked arxiv 0509252

    .. math::
        g_x = g_1 \sin\left(2\phi\right)-g_2\cos\left(2\phi\right)

    For extended descriptions of parameters, see `compute_shear()` documentation.
    """
    return shear1 * np.sin(2.0 * phi) - shear2 * np.cos(2.0 * phi)


def make_radial_profile(
    components,
    angsep,
    angsep_units,
    bin_units,
    bins=10,
    components_error=None,
    error_model="ste",
    include_empty_bins=False,
    return_binnumber=False,
    cosmo=None,
    z_lens=None,
    validate_input=True,
    weights=None,
    coordinate_system=None,
    empty_bins_value=np.nan,
):
    r"""Compute the angular profile of given components

    We assume that the cluster object contains information on the cross and
    tangential shears or ellipticities and angular separation of the source galaxies

    This function can be called in two ways:

    1. Pass explict arguments::

        make_radial_profile([component1, component2], distances, 'radians')

    2. Call it as a method of a GalaxyCluster instance and specify `bin_units`:

        cluster.make_radial_profile('Mpc')

    Parameters
    ----------
    components: list of arrays
        List of arrays to be binned in the radial profile
    angsep: array
        Transvesal distances between the sources and the lens
    angsep_units : str
        Units of the calculated separation of the source galaxies
        Allowed Options = ["radians"]
    bin_units : str
        Units to use for the radial bins of the radial profile
        Allowed Options = ["radians", deg", "arcmin", "arcsec", kpc", "Mpc"]
        (letter case independent)
    bins : array_like, optional
        User defined bins to use for the shear profile. If a list is provided, use that as
        the bin edges. If a scalar is provided, create that many equally spaced bins between
        the minimum and maximum angular separations in bin_units. If nothing is provided,
        default to 10 equally spaced bins.
    components_error: list of arrays, None
        List of errors for input arrays
    error_model : str, optional
        Statistical error model to use for y uncertainties. (letter case independent)
            `ste` - Standard error [=std/sqrt(n) in unweighted computation] (Default).
            `std` - Standard deviation.
    include_empty_bins: bool, optional
        Also include empty bins in the returned table
    return_binnumber: bool, optional
        Also returns the indices of the bins for each object
    cosmo : CLMM.Cosmology
        CLMM Cosmology object to convert angular separations to physical distances
    z_lens: array, optional
        Redshift of the lens
    validate_input: bool
        Validade each input argument
    weights: array-like, optional
        Array of individual galaxy weights. If specified, the radial binned profile is
        computed using a weighted average
    coordinate_system: str, optional
        Coordinate system of the ellipticity components. Must be either 'celestial' or
        euclidean'. See https://doi.org/10.48550/arXiv.1407.7676 section 5.1 for more details.
        Default is 'euclidean'.
    empty_bins_value: float, None
        Values to be assigned to empty bins.

    Returns
    -------
    profile : GCData
        Output table containing the radius grid points, the profile of the components `p_i`,
        errors `p_i_err` and number of sources. The errors are defined as the standard errors in
        each bin.
    binnumber: 1-D ndarray of ints, optional
        Indices of the bins (corresponding to `xbins`) in which each value
        of `xvals` belongs.  Same length as `yvals`.  A binnumber of `i` means the
        corresponding value is between (xbins[i-1], xbins[i]).

    Notes
    -----
    This is an example of a place where the cosmology-dependence can be sequestered to another
    module.
    """
    # pylint: disable-msg=too-many-locals
    if not coordinate_system:
        warnings.warn("coordinate_system not set, defaulting to 'euclidean'")
        coordinate_system = "euclidean"
    if validate_input:
        validate_argument(locals(), "angsep", "float_array")
        validate_argument(locals(), "angsep_units", str)
        validate_argument(locals(), "bin_units", str)
        validate_argument(locals(), "include_empty_bins", bool)
        validate_argument(locals(), "return_binnumber", bool)
        validate_argument(locals(), "z_lens", "float_array", none_ok=True)
        comp_dict = {f"components[{i}]": comp for i, comp in enumerate(components)}
        arguments_consistency(components, names=comp_dict.keys(), prefix="Input components")
        for component in comp_dict:
            validate_argument(comp_dict, component, "float_array")
        _validate_coordinate_system(locals(), "coordinate_system")
    # Check to see if we need to do a unit conversion
    if angsep_units is not bin_units:
        source_seps = convert_units(angsep, angsep_units, bin_units, redshift=z_lens, cosmo=cosmo)
    else:
        source_seps = angsep
    # Make bins if they are not provided
    if not hasattr(bins, "__len__"):
        bins = make_bins(np.min(source_seps), np.max(source_seps), bins)
    # Create output table
    profile_table = GCData(
        [bins[:-1], np.zeros(len(bins) - 1), bins[1:]],
        names=("radius_min", "radius", "radius_max"),
        meta={"bin_units": bin_units, "coordinate_system": coordinate_system},  # Add metadata
    )
    # Compute the binned averages and associated errors
    for i, component in enumerate(components):
        r_avg, comp_avg, comp_err, nsrc, binnumber, wts_sum = compute_radial_averages(
            source_seps,
            component,
            xbins=bins,
            yerr=None if components_error is None else components_error[i],
            error_model=error_model,
            weights=weights,
        )
        profile_table[f"p_{i}"] = comp_avg
        profile_table[f"p_{i}_err"] = comp_err
    profile_table["radius"] = r_avg
    profile_table["n_src"] = nsrc
    profile_table["weights_sum"] = wts_sum
    # return empty bins?
    if not include_empty_bins:
        profile_table = profile_table[nsrc >= 1]
    else:
        for col in (
            "radius",
            *(_col for i in range(len(components)) for _col in (f"p_{i}", f"p_{i}_err")),
        ):
            profile_table[col][nsrc < 1] = empty_bins_value
    if return_binnumber:
        return profile_table, binnumber
    return profile_table


def not_nan_average(values, axis=0, weights=None):
    """Computes averages using only not nan values

    Parameters
    ----------
    values: nd array
        Values to be averaged
    axis: int
        axis to make the average on
    weighs: nd array
        Weights, must have same shape as values

    Returns
    -------
    array
        Averaged values
    """
    _values = np.copy(values)
    _weights = np.ones(_values.shape) if weights is None else np.copy(weights)
    _values[np.isnan(values)] = 0
    _weights[np.isnan(values)] = 0
    return np.average(_values, axis=axis, weights=_weights)


def make_stacked_radial_profile(angsep, weights, components):
    """Compute stacked profile, and mean separation distances.

    Parameters
    ----------
    angsep: 2d array
        Transvesal distances corresponding to each object with shape `n_obj, n_rad_bins`.
    weights: 2d array
        Weights corresponding to each objects with shape `n_obj, n_rad_bins`.
    components: list of 2d arrays
        List of 2d properties of each array to be stacked with shape
        `n_components, n_obj, n_rad_bins`.

    Returns
    -------
    staked_angsep: array
        Mean transversal distance in each radial bin.
    stacked_components: list of arrays
        List of stacked components.
    """
    # rm nan values
    staked_angsep = not_nan_average(angsep, axis=0, weights=None)
    stacked_components = [
        not_nan_average(component, axis=0, weights=weights) for component in components
    ]
    return staked_angsep, stacked_components



"""General utility functions that are used in multiple modules"""

import warnings

import numpy as np
from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import interp1d
from scipy.stats import binned_statistic


def compute_weighted_bin_sum(xvals, yvals, xbins, weights):
    """Add `yvals * weights` in `xbins`.

    Parameters
    ----------
    xvals : array_like
        Values to be binned
    yvals : array_like
        Values to compute statistics on
    xbins: array_like
        Bin edges to sort into
    weights: array_like, None, optional
        Weights for sum.

    Returns
    -------
    numpy.ndarray
        Sum of `yvals * weights` in `xbins`.
    """
    return binned_statistic(xvals, yvals * weights, statistic="sum", bins=xbins)[0]


def compute_radial_averages(xvals, yvals, xbins, yerr=None, error_model="ste", weights=None):
    """Given a list of `xvals`, `yvals` and `xbins`, sort into bins. If `xvals` or `yvals`
    contain non-finite values, these are filtered.

    Parameters
    ----------
    xvals : array_like
        Values to be binned
    yvals : array_like
        Values to compute statistics on
    xbins: array_like
        Bin edges to sort into
    yerr : array_like, None, optional
        Errors of `yvals`. Default: None
    error_model : str, optional
        Statistical error model to use for y uncertainties. (letter case independent)

            * 'ste' - Standard error [=std/sqrt(n) in unweighted computation] (Default).
            * 'std' - Standard deviation.

    weights: array_like, None, optional
        Weights for averages. Default: None


    Returns
    -------
    mean_x : numpy.ndarray
        Mean x value in each bin
    mean_y : numpy.ndarray
        Mean y value in each bin
    err_y: numpy.ndarray
        Error on the mean y value in each bin. Specified by `error_model`
    num_objects : numpy.ndarray
        Number of objects in each bin
    binnumber: 1-D ndarray of ints
        Indices of the bins (corresponding to `xbins`) in which each value
        of `xvals` belongs.  Same length as `yvals`.  A binnumber of `i` means the
        corresponding value is between (xbins[i-1], xbins[i]).
    wts_sum: numpy.ndarray
        Sum of individual weights in each bin.
    """
    # make case independent
    error_model = error_model.lower()
    # binned_statics throus an error in case of non-finite values, so filtering those out
    filt = np.isfinite(xvals) * np.isfinite(yvals)
    xfilt, yfilt = np.array(xvals)[filt], np.array(yvals)[filt]
    # normalize weights (and computers binnumber)
    wts = np.ones(xfilt.size) if weights is None else np.array(weights, dtype=float)[filt]
    wts_sum, binnumber = binned_statistic(xfilt, wts, statistic="sum", bins=xbins)[:3:2]
    objs_in_bins = (binnumber > 0) * (binnumber <= wts_sum.size)  # mask for binnumber in range
    wts[objs_in_bins] *= 1.0 / wts_sum[binnumber[objs_in_bins] - 1]  # norm weights in each bin
    # means
    mean_x = compute_weighted_bin_sum(xfilt, xfilt, xbins, wts)
    mean_y = compute_weighted_bin_sum(xfilt, yfilt, xbins, wts)
    # errors
    data_yerr2 = (
        0
        if yerr is None
        else compute_weighted_bin_sum(xfilt, np.array(yerr)[filt] ** 2, xbins, wts**2)
    )
    stat_yerr2 = compute_weighted_bin_sum(xfilt, yfilt**2, xbins, wts) - mean_y**2
    if error_model == "ste":
        # sum(wts^2)=1/n for not weighted
        stat_yerr2 *= compute_weighted_bin_sum(xfilt, wts, xbins, wts)
    elif error_model != "std":
        raise ValueError(f"{error_model} not supported err model for binned stats")
    err_y = np.sqrt(stat_yerr2 + data_yerr2)
    # number of objects
    num_objects = np.histogram(xfilt, xbins)[0]
    return mean_x, mean_y, err_y, num_objects, binnumber, wts_sum


def make_bins(rmin, rmax, nbins=10, method="evenwidth", source_seps=None):
    """Define bin edges

    Parameters
    ----------
    rmin : float, None
        Minimum bin edges wanted. If None, min(`source_seps`) is used.
    rmax : float, None
        Maximum bin edges wanted. If None, max(`source_seps`) is used.
    nbins : float, optional
        Number of bins you want to create, default to 10.
    method : str, optional
        Binning method to use (letter case independent):

            * 'evenwidth' - Default, evenly spaced bins between `rmin` and `rmax`
            * 'evenlog10width' - Logspaced bins with even width in log10 between `rmin` and `rmax`
            * 'equaloccupation' - Bins with equal occupation numbers

    source_seps : array_like, None, optional
        Radial distance of source separations. Needed if `method='equaloccupation'`. Default: None

    Returns
    -------
    binedges: numpy.ndarray
        array with `nbins` +1 elements that defines bin edges
    """
    # make case independent
    method = method.lower()
    if method == "equaloccupation":
        if source_seps is None:
            raise ValueError(f"Binning method '{method}' requires source separations array")
        seps = np.array(source_seps)
        rmin = seps.min() if rmin is None else rmin
        rmax = seps.max() if rmax is None else rmax
        # Need to filter source_seps to only keep galaxies in the [rmin, rmax] with a mask
        mask = (seps >= rmin) * (seps <= rmax)
        binedges = np.percentile(seps[mask], tuple(np.linspace(0, 100, nbins + 1, endpoint=True)))
    else:
        # Check consistency
        if (rmin > rmax) or (rmin < 0.0) or (rmax < 0.0):
            raise ValueError(f"Invalid bin endpoints in make_bins, {rmin} {rmax}")
        if (nbins <= 0) or not isinstance(nbins, int):
            raise ValueError(f"Invalid nbins={nbins}. Must be integer greater than 0.")

        if method == "evenwidth":
            binedges = np.linspace(rmin, rmax, nbins + 1, endpoint=True)
        elif method == "evenlog10width":
            binedges = np.logspace(np.log10(rmin), np.log10(rmax), nbins + 1, endpoint=True)
        else:
            raise ValueError(f"Binning method '{method}' is not currently supported")

    return binedges


def _draw_random_points_from_distribution(xmin, xmax, nobj, dist_func, xstep=0.001):
    """Draw random points with a given distribution.

    Uses a sampling technique found in Numerical Recipes in C, Chap 7.2: Transformation Method.

    Parameters
    ----------
    xmin : float
        The minimum source redshift allowed.
    xmax : float, optional
        If source redshifts are drawn, the maximum source redshift
    nobj : float
        Number of galaxies to generate
    dist_func : function
        Function of the required distribution
    xstep : float, optional
        Size of the step to interpolate the culmulative distribution. Default: 0.001

    Returns
    -------
    numpy.ndarray
        Random points with dist_func distribution
    """
    steps = int((xmax - xmin) / xstep) + 1
    xdomain = np.linspace(xmin, xmax, steps)
    # Cumulative probability function of the redshift distribution
    # probdist = np.vectorize(lambda zmax: integrate.quad(dist_func, xmin, zmax)[0])(xdomain)
    probdist = dist_func(xdomain, is_cdf=True) - dist_func(xmin, is_cdf=True)
    # Get random values for probdist
    uniform_deviate = np.random.uniform(probdist.min(), probdist.max(), nobj)
    return interp1d(probdist, xdomain, kind="linear")(uniform_deviate)


def _draw_random_points_from_tab_distribution(x_tab, pdf_tab, nobj=1, xmin=None, xmax=None):
    """Draw random points from a tabulated distribution.

    Parameters
    ----------
    x_tab : array-like
        Values for which the tabulated pdf is provided
    pdf_tab : array-like
        Value of the pdf at the x_tab locations
    nobj : int, optional
        Number of random samples to generate. Default is 1.
    xmin : float, optional
        Lower bound to draw redshift. Default is the min(x_tab)
    xmax : float, optional
        Upper bound to draw redshift. Default is the max(x_tab)

    Returns
    -------
    samples : numpy.ndarray
        Random points following the pdf_tab distribution
    """
    x_tab = np.array(x_tab)
    pdf_tab = np.array(pdf_tab)
    # cdf = np.array([simps(pdf_tab[:j], x_tab[:j]) for j in range(1, len(x_tab)+1)])
    cdf = cumulative_trapezoid(pdf_tab, x_tab, initial=0)
    # Normalise it
    cdf /= cdf.max()
    cdf_xmin, cdf_xmax = 0.0, 1.0
    # Interpolate cdf at xmin and xmax
    if xmin or xmax:
        cdf_interp = interp1d(x_tab, cdf, kind="linear")
        if xmin is not None:
            if xmin < x_tab.min():
                warnings.warn(
                    "`xmin` is less than the minimum value of `x_tab`. "
                    + f"Using min(x_tab)={x_tab.min()} instead."
                )
            else:
                cdf_xmin = cdf_interp(xmin)
        if xmax is not None:
            if xmax > x_tab.max():
                warnings.warn(
                    "`xmax` is greater than the maximum value of `x_tab`. "
                    + f"Using max(x_tab)={x_tab.max()} instead."
                )
            else:
                cdf_xmax = cdf_interp(xmax)
    # Interpolate the inverse CDF
    inv_cdf = interp1d(cdf, x_tab, kind="linear", bounds_error=False, fill_value=0.0)
    # Finally generate sample from uniform distribution and
    # get the corresponding samples
    samples = inv_cdf(np.random.random(nobj) * (cdf_xmax - cdf_xmin) + cdf_xmin)
    return samples


def gaussian(value, mean, scatter):
    """Normal distribution.

    Parameters
    ----------
    value : array-like
        Values for which to evaluate gaussian.
    mean : float
        Mean value of normal distribution
    scatter : float
        Scatter of normal distribution

    Returns
    -------
    numpy.ndarray
        Gaussian values at `value`
    """
    return np.exp(-0.5 * (value - mean) ** 2 / scatter**2) / np.sqrt(2 * np.pi * scatter**2)



"""General utility functions that are used in multiple modules"""

from . import constants, plotting, redshift_distributions
from .beta_lens import (
    compute_beta,
    compute_beta_s,
    compute_beta_s_func,
    compute_beta_s_mean_from_distribution,
    compute_beta_s_mean_from_weights,
    compute_beta_s_square_mean_from_distribution,
    compute_beta_s_square_mean_from_weights,
)
from .boost import (
    boost_models,
    compute_nfw_boost,
    compute_powerlaw_boost,
    correct_with_boost_model,
    correct_with_boost_values,
)
from .ellipticity import (
    build_ellipticities,
    compute_lensed_ellipticity,
    convert_shapes_to_epsilon,
)
from .redshift_tools import _integ_pzfuncs, compute_for_good_redshifts
from .statistic import (
    _draw_random_points_from_distribution,
    _draw_random_points_from_tab_distribution,
    compute_radial_averages,
    compute_weighted_bin_sum,
    gaussian,
    make_bins,
)
from .units import convert_units
from .validation import (
    DiffArray,
    _patch_rho_crit_to_cd2018,
    _validate_coordinate_system,
    _validate_dec,
    _validate_is_deltasigma_sigma_c,
    _validate_ra,
    arguments_consistency,
    validate_argument,
)



"""A collection of scripts that can be used to plot the various quantities that CLMM models."""

# Set mpl backend run plots on github actions
import os

import matplotlib as mpl

# pylint: disable=wrong-import-position
if os.environ.get("DISPLAY", "") == "test":
    print("no display found. Using non-interactive Agg backend")
    mpl.use("Agg")
import matplotlib.pyplot as plt


def plot_profiles(
    rbins,
    tangential_component,
    tangential_component_error,
    cross_component,
    cross_component_error,
    r_units=None,
    xscale="linear",
    yscale="linear",
    tangential_component_label="Tangential component",
    cross_component_label="Cross component",
):
    """Plot shear profiles

    Parameters
    ----------
    rbins: array_like
        The centers of the radial bins that was used to compute the shears.
    tangential_component: array_like
        The tangential component at the radii of `rbins`
    tangential_component_error: array_like
        The uncertainty in the tangential component
    cross_component: array_like
        The cross component at the radii of `rbins`
    cross_component_error: array_like
        The uncertainty in the cross component
    r_units: str
        Units of `rbins` for x-axis label
    xscale:
        matplotlib.pyplot.xscale parameter to set x-axis scale (e.g. to logarithmic axis)
    yscale:
        matplotlib.pyplot.yscale parameter to set y-axis scale (e.g. to logarithmic axis)

    Returns
    -------
    fig:
        The matplotlib figure object that has been plotted to.
    axes:
        The matplotlib axes object that has been plotted to.
    """
    # Plot the tangential shears
    fig, axes = plt.subplots()
    axes.errorbar(
        rbins,
        tangential_component,
        yerr=tangential_component_error,
        fmt="bo-",
        label=tangential_component_label,
        lw=0.2,
        markersize=1.5,
        elinewidth=0.5,
        capthick=3,
    )
    # Plot the cross shears
    axes.errorbar(
        rbins,
        cross_component,
        yerr=cross_component_error,
        fmt="ro-",
        label=cross_component_label,
        lw=0.2,
        markersize=1.5,
        elinewidth=0.5,
        capthick=3,
    )
    # format
    axes.set_xscale(xscale)
    axes.set_yscale(yscale)
    axes.legend()
    axes.set_xlabel(f"Radius [{r_units}]")
    axes.set_ylabel(rf"{tangential_component_label} | {cross_component_label}")

    return fig, axes



"""@file ccl.py
Modeling using CCL
"""

import pyccl as ccl

# Functions to model halo profiles
from packaging.version import parse

from ..cosmology.ccl import CCLCosmology
from ..utils import _patch_rho_crit_to_cd2018

# Check which versions of ccl are currently supported
from . import _ccl_supported_versions
from .parent_class import CLMModeling

if (
    parse(ccl.__version__) < parse(_ccl_supported_versions.VMIN)
    or parse(ccl.__version__).major > parse(_ccl_supported_versions.VMAX).major
):
    raise EnvironmentError(
        f"Current CCL version ({ccl.__version__}) not supported by CLMM. "
        f"It must be between {_ccl_supported_versions.VMIN} and {_ccl_supported_versions.VMAX}."
    )


class CCLCLMModeling(CLMModeling):
    r"""Object with functions for halo mass modeling

    Attributes
    ----------
    backend: str
        Name of the backend being used
    massdef : str
        Profile mass definition ("mean", "critical", "virial" - letter case independent)
    delta_mdef : int
        Mass overdensity definition.
    halo_profile_model : str
        Profile model parameterization ("nfw", "einasto", "hernquist" - letter case independent)
    cosmo: Cosmology
        Cosmology object
    hdpm: Object
        Backend object with halo profiles
    mdef_dict: dict
        Dictionary with the definitions for mass
    hdpm_dict: dict
        Dictionary with the definitions for profile
    mdef: ccl.halos.MassDef, None
        Internal MassDef object
    conc: ccl.halos.ConcentrationConstant, None
        Internal ConcentrationConstant object
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        massdef="mean",
        delta_mdef=200,
        halo_profile_model="nfw",
        validate_input=True,
    ):
        CLMModeling.__init__(self, validate_input)
        # Update class attributes
        self.backend = "ccl"
        self.mdef_dict = {
            "mean": "matter",
            "critical": "critical",
            "virial": "critical",
        }
        self.hdpm_dict = {
            "nfw": ccl.halos.HaloProfileNFW,
            "einasto": ccl.halos.HaloProfileEinasto,
            "hernquist": ccl.halos.HaloProfileHernquist,
        }
        self.cosmo_class = CCLCosmology
        self.hdpm_opts = {
            "nfw": {
                "truncated": False,
                "projected_analytic": True,
                "cumul2d_analytic": True,
            },
            "einasto": {"truncated": False},
            "hernquist": {
                "truncated": False,
                "projected_analytic": True,
                "cumul2d_analytic": True,
            },
        }
        self.cor_factor = _patch_rho_crit_to_cd2018(ccl.physical_constants.RHO_CRITICAL)
        self.__mdelta_cor = 0.0  ## mass with corretion for input
        # self.hdpm_opts['einasto'].update({'alpha': 0.25}) # same as NC default

        # Set halo profile and cosmology
        self.mdef = None
        self.conc = None
        self.set_halo_density_profile(halo_profile_model, massdef, delta_mdef)
        self.set_cosmo(None)

    def _set_projected_quad(self, use_projected_quad):
        if hasattr(self.hdpm, "projected_quad"):
            self.hdpm_opts["einasto"]["projected_quad"] = use_projected_quad
            self._update_halo_density_profile()
        else:
            raise NotImplementedError("projected_quad is not available on this version of CCL.")

    # Functions implemented by child class

    def _update_halo_density_profile(self):
        """updates halo density profile with set internal properties"""
        # prepare mdef object
        if self.massdef == "virial":
            self.mdef = ccl.halos.MassDef("vir", self.mdef_dict[self.massdef])
        else:
            self.mdef = ccl.halos.MassDef(self.delta_mdef, self.mdef_dict[self.massdef])
        # setting concentration (also updates hdpm)
        self.cdelta = self.cdelta if self.hdpm else 4.0  # ccl always needs an input concentration

    def _get_concentration(self):
        """get concentration"""
        return self.conc.c

    def _get_mass(self):
        """get mass"""
        return self.__mdelta_cor * self.cor_factor

    def _set_concentration(self, cdelta):
        """set concentration. Also sets/updates hdpm"""
        self.conc = ccl.halos.ConcentrationConstant(c=cdelta, mass_def=self.mdef)
        self.hdpm = self.hdpm_dict[self.halo_profile_model](
            concentration=self.conc, mass_def=self.mdef, **self.hdpm_opts[self.halo_profile_model]
        )
        self.hdpm.update_precision_fftlog(padding_lo_fftlog=1e-4, padding_hi_fftlog=1e3)

    def _set_mass(self, mdelta):
        """set mass"""
        self.__mdelta_cor = mdelta / self.cor_factor

    def _set_einasto_alpha(self, alpha):
        if alpha is None:
            self.hdpm.update_parameters(alpha="cosmo")
        else:
            self.hdpm.update_parameters(alpha=alpha)

    def _get_einasto_alpha(self, z_cl=None):
        """get the value of the Einasto slope"""
        # pylint: disable=protected-access
        if self.hdpm.alpha != "cosmo":
            a_cl = 1  # a_cl does not matter in this case
        else:
            a_cl = self.cosmo.get_a_from_z(z_cl)
        return self.hdpm._get_alpha(self.cosmo.be_cosmo, self.__mdelta_cor, a_cl)

    def _get_delta_mdef_virial(self, z_cl):
        "Gets the virial overdensity"
        a_cl = self.cosmo.get_a_from_z(z_cl)
        # This fuction can only return the virial overdensity
        return self.mdef.get_Delta_vir(self.cosmo.be_cosmo, a_cl)

    def _eval_3d_density(self, r3d, z_cl):
        """eval 3d density"""
        return self._call_ccl_profile_lens(self.hdpm.real, r3d, z_cl, ndim=3)

    def _eval_surface_density(self, r_proj, z_cl):
        """eval surface density"""
        return self._call_ccl_profile_lens(self.hdpm.projected, r_proj, z_cl)

    def _eval_mean_surface_density(self, r_proj, z_cl):
        """eval mean surface density"""
        return self._call_ccl_profile_lens(self.hdpm.cumul2d, r_proj, z_cl)

    def _eval_excess_surface_density(self, r_proj, z_cl):
        """eval excess surface density"""
        return self._eval_mean_surface_density(r_proj, z_cl) - self._eval_surface_density(
            r_proj, z_cl
        )

    def _eval_convergence_core(self, r_proj, z_cl, z_src):
        """eval convergence"""
        return self._call_ccl_profile_lens_src(self.hdpm.convergence, r_proj, z_cl, z_src)

    def _eval_tangential_shear_core(self, r_proj, z_cl, z_src):
        """eval tangential shear"""
        return self._call_ccl_profile_lens_src(self.hdpm.shear, r_proj, z_cl, z_src)

    def _eval_reduced_tangential_shear_core(self, r_proj, z_cl, z_src):
        """eval reduced tangential shear with all background sources at the same plane"""
        return self._call_ccl_profile_lens_src(self.hdpm.reduced_shear, r_proj, z_cl, z_src)

    def _eval_magnification_core(self, r_proj, z_cl, z_src):
        """eval magnification"""
        return self._call_ccl_profile_lens_src(self.hdpm.magnification, r_proj, z_cl, z_src)

    # Helper functions unique to this class

    def _call_ccl_profile_lens(self, ccl_hdpm_func, radius, z_lens, ndim=2):
        """call ccl profile functions that depend on the lens only"""
        a_lens = self.cosmo.get_a_from_z(z_lens)

        return (
            ccl_hdpm_func(
                self.cosmo.be_cosmo,
                radius / a_lens,
                self.__mdelta_cor,
                a_lens,
            )
            * self.cor_factor
            / a_lens**ndim
        )

    def _call_ccl_profile_lens_src(self, ccl_hdpm_func, radius, z_lens, z_src):
        """call ccl profile functions that depend on the lens and the sources"""
        a_lens = self.cosmo.get_a_from_z(z_lens)
        a_src = self.cosmo.get_a_from_z(z_src)

        return ccl_hdpm_func(
            self.cosmo.be_cosmo,
            radius / a_lens,
            self.__mdelta_cor,
            a_lens=a_lens,
            a_source=a_src,
        )


Cosmology = CCLCosmology
Modeling = CCLCLMModeling



"""
Define the custom data type
"""

import warnings
from collections import OrderedDict
from copy import deepcopy

import numpy as np
import qp
from astropy.table import Table as APtable

from .utils import _validate_coordinate_system


class GCMetaData(OrderedDict):
    r"""Object to store metadata, it always has a cosmo and coordinate_system keys with protective
    changes

    Attributes
    ----------
    protected: bool
        Protect cosmo key
    OrderedDict attributes
    """

    def __init__(self, *args, **kwargs):
        OrderedDict.__init__(self, *args, **kwargs)
        if "coordinate_system" not in self:
            warnings.warn("coordinate_system not set, defaulting to 'euclidean'.")
            coordinate_system = "euclidean"
        else:
            coordinate_system = kwargs["coordinate_system"]
        _validate_coordinate_system(locals(), "coordinate_system")
        OrderedDict.__setitem__(self, "coordinate_system", coordinate_system)
        if "cosmo" not in self:
            OrderedDict.__setitem__(self, "cosmo", None)

    def __setitem__(self, item, value):
        if item == "cosmo" and self.get("cosmo", None):
            raise ValueError(
                "cosmo must be changed via update_cosmo or update_cosmo_ext_valid method"
            )
        if item == "coordinate_system" and self.get("coordinate_system", None):
            raise ValueError(
                "coordinate_system must be changed via update_coordinate_system method"
            )
        OrderedDict.__setitem__(self, item, value)

    def __getitem__(self, item):
        """
        Make class accept all letter casings
        """
        if isinstance(item, str):
            item = {n.lower(): n for n in self.keys()}[item.lower()]
        out = OrderedDict.__getitem__(self, item)
        return out

    def __deepcopy__(self, memo):
        """
        Make sure that operations dependent on deepcopy avoid initializing an empty GCMetaData
        and then setting cosmo and coordinate_system keys incorrectly.
        """
        items = {key: deepcopy(value, memo) for key, value in self.items()}
        return GCMetaData(**items)

    def __reduce__(self):
        """
        Make sure that metadata can be pickled without calling __setitem__ for cosmo and
        coordinate_system keys.
        """
        return (self.__class__.__new__, (self.__class__,), dict(self))

    def __setstate__(self, state):
        """
        Make sure that metadata can be unpickled without calling __setitem__ for cosmo and
        coordinate_system keys.
        """
        self.clear()
        self.update(state)

    def copy(self):
        return GCMetaData(**self)


class GCData(APtable):
    """
    GCData: A data object for gcdata. Right now it behaves as an astropy table, with the following
    modifications: `__getitem__` is case independent;
    The attribute .meta['cosmo'] is protected and
    can only be changed via update_cosmo or update_cosmo_ext_valid methods;

    Parameters
    ----------
    meta: dict
        Dictionary with metadata for this object

    Same as astropy tables
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        *args, **kwargs: Same used for astropy tables
        """
        APtable.__init__(self, *args, **kwargs)
        if not ("copy" in kwargs or "copy_indices" in kwargs or "masked" in kwargs):
            metakwargs = kwargs.get("meta", {})
            self.meta = GCMetaData(**metakwargs)
        # this attribute is set when source galaxies have p(z)
        self.pzpdf_info = {
            "type": None,
            "unpack_quantile_zbins_limits": (0, 5, 501),
        }

    def _str_colnames(self):
        """Colnames in comma separated str"""
        return ", ".join(self.colnames)

    def _str_meta_(self):
        """Metadata in comma separated str"""
        return ", ".join([f"{key}={value!r}" for key, value in self.meta.items()])

    def _str_pzpdf_info(self):
        out = self.pzpdf_info["type"]
        if out is not None:
            if out == "shared_bins":
                default_cfg = np.get_printoptions()  # keep default values
                np.set_printoptions(edgeitems=5, threshold=10)
                out += " " + str(np.round(self.pzpdf_info.get("zbins"), 2))
                np.set_printoptions(**default_cfg)
            elif out == "quantiles":
                np.set_printoptions(formatter={"float": "{0:g}".format}, edgeitems=3, threshold=6)
                out += " " + str(self.pzpdf_info["quantiles"])
                out += " - unpacked with zgrid : " + str(
                    self.pzpdf_info["unpack_quantile_zbins_limits"]
                )
        return out

    def __repr__(self):
        """Generates string for repr(GCData)"""
        description = [self._str_meta_(), "columns: " + self._str_colnames()]
        if self.pzpdf_info["type"]:
            description.append(f"pzpdf: {self.pzpdf_info['type']}")
        return f'{self.__class__.__name__}({", ".join(description)})'

    def __str__(self):
        """Generates string for print(GCData)"""
        out = [
            f"{self.__class__.__name__}",
            f"> defined by: {self._str_meta_()}",
            f"> with columns: {self._str_colnames()}",
            f"> {len(self)} objects",
            f"{APtable.__str__(self)}",
        ]
        if self.pzpdf_info["type"]:
            out.insert(3, f"> and pzpdf: {self._str_pzpdf_info()}")
        return "\n".join(out)

    def _html_table(self):
        """Get html table for display"""
        return "</i>".join(APtable._repr_html_(self).split("</i>")[1:])

    def _repr_html_(self):
        """Generates string for display(GCData)"""
        out = [
            f"<b>{self.__class__.__name__}</b>",
            f"<br> <b>defined by:</b> {self._str_meta_()}",
            f"<br> <b>with columns:</b> {self._str_colnames()}",
            f"<br> {len(self)} objects",
            f"<br> {self._html_table()}",
        ]
        if self.pzpdf_info["type"]:
            out.insert(3, f"<br> <b>and pzpdf:</b> {self._str_pzpdf_info()}")
        return "\n".join(out)

    def __getitem__(self, item):
        """
        Makes sure GCData keeps its properties after [] operations are used.
        It also makes all letter casings accepted

        Returns
        -------
        GCData
            Data with [] operations applied
        """
        if isinstance(item, str):
            name_dict = {n.lower(): n for n in self.colnames}
            item = item.lower()
            item = ",".join([name_dict[i] for i in item.split(",")])
        out = APtable.__getitem__(self, item)
        # sub cols or sub rows
        if not isinstance(item, (str, int, np.int64)):
            out.pzpdf_info = self.pzpdf_info
        return out

    def update_info_ext_valid(self, key, gcdata, ext_value, overwrite=False):
        r"""Updates cosmo metadata if the same as in gcdata

        Parameters
        ----------
        key: str
            Name of key to compare and update.
        gcdata: GCData
            Table to check if same cosmology and ensemble bins.
        ext_value:
            Value to be compared to.
        overwrite: bool
            Overwrites the current metadata. If false raises Error when values are different.

        Returns
        -------
        None
        """
        if key == "coordinate_system":
            _validate_coordinate_system(locals(), "ext_value")
        if ext_value:
            in_value = gcdata.meta[key]
            if in_value and in_value != ext_value:
                if overwrite:
                    warnings.warn(
                        f"input '{key}' ({ext_value}) overwriting gcdata '{key}' ({in_value})"
                    )
                else:
                    raise ValueError(
                        f"input '{key}' ({ext_value}) differs from gcdata '{key}' ({in_value})"
                    )
            OrderedDict.__setitem__(self.meta, key, ext_value)

    def update_cosmo_ext_valid(self, gcdata, cosmo, overwrite=False):
        r"""Updates cosmo metadata if the same as in gcdata

        Parameters
        ----------
        gcdata: GCData
            Table to check if same cosmology
        cosmo: clmm.Cosmology
            Cosmology
        overwrite: bool
            Overwrites the current cosmo metadata. If false raises Error when cosmologies are
            different.

        Returns
        -------
        None
        """
        cosmo_desc = cosmo.get_desc() if cosmo else None
        self.update_info_ext_valid("cosmo", gcdata, cosmo_desc, overwrite=overwrite)

    def update_cosmo(self, cosmo, overwrite=False):
        r"""Updates cosmo metadata if not present

        Parameters
        ----------
        cosmo: clmm.Cosmology
            Cosmology
        overwrite: bool
            Overwrites the current cosmo metadata. If false raises Error when cosmologies are
            different.

        Returns
        -------
        None
        """
        self.update_cosmo_ext_valid(self, cosmo, overwrite=overwrite)

    def update_coordinate_system(self, coordinate_system, ellipticity_columns=()):
        r"""Updates coordinate_system metadata and converts ellipticity

        Parameters
        ----------
        coordinate_system : str
            Coordinate system of the ellipticity components. Must be either 'celestial' or
            euclidean'. See https://doi.org/10.48550/arXiv.1407.7676 section 5.1 for more details.
        ellipticity_columns : tuple
            Name of the columns holding the ellipticity components. Passing names not corresponding
            to these columns will have undesired effects.

        Returns
        -------
        None
        """
        if coordinate_system == self.meta["coordinate_system"]:
            warnings.warn(
                f"coordinate_system {coordinate_system} is the same as in data. No changes made."
            )
            return

        missing_columns = list(filter(lambda col: col not in self.colnames, ellipticity_columns))
        if len(missing_columns) > 0:
            raise ValueError(f"Components {missing_columns} not found in data. No changes made")

        self.update_info_ext_valid("coordinate_system", self, coordinate_system, overwrite=True)
        for col in ellipticity_columns:
            self[col] *= -1

        if ellipticity_columns:
            warnings.warn(f"Columns {ellipticity_columns} have been updated.")
        else:
            warnings.warn(
                "No ellipticity columns provided. Coordinate system has been updated but "
                "no changes were made to the data."
            )

    def has_pzpdfs(self):
        """Get pzbins and pzpdfs of galaxies

        Returns
        -------
        pzbins : array
            zbins of each object in data
        pzpdfs : array
            PDF of each object in data
        """
        pzpdf_type = self.pzpdf_info["type"]
        if pzpdf_type is None:
            return False
        if pzpdf_type == "shared_bins":
            return ("zbins" in self.pzpdf_info) and ("pzpdf" in self.columns)
        if pzpdf_type == "individual_bins":
            return ("pzbins" in self.columns) and ("pzpdf" in self.columns)
        if pzpdf_type == "quantiles":
            return ("quantiles" in self.pzpdf_info) and ("pzquantiles" in self.columns)
        raise NotImplementedError(f"PDF use '{pzpdf_type}' not implemented.")

    def get_pzpdfs(self):
        """Get pzbins and pzpdfs of galaxies

        Returns
        -------
        pzbins : array
            zbins of PDF. 1D if `shared_bins` or `quantiles`.
            zbins of each object in data if `individual_bins`.
        pzpdfs : array
            PDF of each object in data

        Notes
        -----
        If pzpdf type is quantiles, a pdf will be unpacked on a grid contructed with
        `np.linspace(*self.pzpdf_info["unpack_quantile_zbins_limits"])`
        """
        pzpdf_type = self.pzpdf_info["type"]
        if pzpdf_type is None:
            raise ValueError("No PDF information stored!")
        if pzpdf_type == "shared_bins":
            pzbins = self.pzpdf_info["zbins"]
            pzpdf = self["pzpdf"]
        elif pzpdf_type == "individual_bins":
            pzbins = self["pzbins"]
            pzpdf = self["pzpdf"]
        elif pzpdf_type == "quantiles":
            pzbins = np.linspace(*self.pzpdf_info["unpack_quantile_zbins_limits"])
            qp_ensemble = qp.Ensemble(
                qp.quant,
                data={
                    "quants": np.array(self.pzpdf_info["quantiles"]),
                    "locs": self["pzquantiles"],
                },
            )
            pzpdf = qp_ensemble.pdf(pzbins)
        else:
            raise NotImplementedError(f"PDF use '{pzpdf_type}' not implemented.")
        return pzbins, pzpdf



"""@file parent_class.py
CLMMCosmology abstract class
"""

# CLMM Cosmology object abstract superclass
import numpy as np

from ..utils import compute_for_good_redshifts, validate_argument
from ..utils.constants import Constants as const


class CLMMCosmology:
    """
    Cosmology object superclass for supporting multiple back-end cosmology objects

    Attributes
    ----------
    backend: str
        Name of back-end used
    be_cosmo: cosmology library
        Cosmology library used in the back-end
    validate_input: bool
        Validade each input argument
    additional_config: dict
        Dictionary with additional (implicit) config that will be used by the class.
    """

    def __init__(self, validate_input=True, **kwargs):
        self.backend = None
        self.be_cosmo = None
        self.validate_input = validate_input
        self.additional_config = {}
        self.set_be_cosmo(**kwargs)

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._get_param(key)
        raise TypeError(f"input must be str, not {type(key)}")

    def __setitem__(self, key, val):
        if isinstance(key, str):
            self._set_param(key, val)
        else:
            raise TypeError(f"key input must be str, not {type(key)}")

    # 1. Functions to be implemented by children classes

    def _init_from_cosmo(self, be_cosmo):
        raise NotImplementedError

    def _init_from_params(self, **kwargs):
        raise NotImplementedError

    def _set_param(self, key, value):
        raise NotImplementedError

    def _get_param(self, key):
        raise NotImplementedError

    def _get_Omega_m(self, z):
        raise NotImplementedError

    def _get_E2(self, z):
        raise NotImplementedError

    def _get_E2Omega_m(self, z):
        raise NotImplementedError

    def _get_rho_c(self, z):
        raise NotImplementedError

    def _eval_da_z1z2_core(self, z1, z2):
        raise NotImplementedError

    def _eval_sigma_crit_core(self, z_len, z_src):
        raise NotImplementedError

    def _eval_linear_matter_powerspectrum(self, k_vals, redshift):
        raise NotImplementedError

    # 2. Functions that can be used by all subclasses

    def _get_rho_m(self, z):
        rhocrit_cd2018 = (3.0e16 * const.PC_TO_METER.value) / (
            8.0 * np.pi * const.GNEWT.value * const.SOLAR_MASS.value
        )
        return rhocrit_cd2018 * (z + 1) ** 3 * self["Omega_m0"] * self["h"] ** 2

    def _eval_da_z1z2(self, z1, z2):
        warning_msg = "\nSome values of z2 are lower than z1." + "\nda = np.nan for those."
        return compute_for_good_redshifts(
            self._eval_da_z1z2_core, z1, z2, np.nan, warning_message=warning_msg
        )

    def _eval_da(self, z):
        return self._eval_da_z1z2(0.0, z)

    def _get_a_from_z(self, z):
        z = np.array(z)
        return 1.0 / (1.0 + z)

    def _get_z_from_a(self, a):
        a = np.array(a)
        return (1.0 / a) - 1.0

    def _eval_sigma_crit(self, z_len, z_src):
        warning_msg = (
            "\nSome source redshifts are lower than the cluster redshift."
            + "\nSigma_crit = np.inf for those galaxies."
        )
        return compute_for_good_redshifts(
            self._eval_sigma_crit_core,
            z_len,
            z_src,
            np.inf,
            z1_arg_name="z_len",
            z2_arg_name="z_src",
            warning_message=warning_msg,
        )

    # 3. Wrapper functions for input validation

    def get_desc(self):
        """
        Returns the Cosmology description.
        """
        return (
            f"{type(self).__name__}(H0={self['H0']}, Omega_dm0={self['Omega_dm0']}, "
            f"Omega_b0={self['Omega_b0']}, Omega_k0={self['Omega_k0']})"
        )

    def init_from_params(self, H0=67.66, Omega_b0=0.049, Omega_dm0=0.262, Omega_k0=0.0):
        """Set the cosmology from parameters

        Parameters
        ----------
        H0: float
            Hubble parameter.
        Omega_b0: float
            Mass density of baryons today.
        Omega_dm0: float
            Mass density of dark matter only (no baryons) today.
        Omega_k0: float
            Mass density of curvature today.
        """
        if self.validate_input:
            validate_argument(locals(), "H0", float, argmin=0)
            validate_argument(locals(), "Omega_b0", float, argmin=0, eqmin=True)
            validate_argument(locals(), "Omega_dm0", float, argmin=0, eqmin=True)
            validate_argument(locals(), "Omega_k0", float, argmin=0, eqmin=True)
        self._init_from_params(H0=H0, Omega_b0=Omega_b0, Omega_dm0=Omega_dm0, Omega_k0=Omega_k0)

    def set_be_cosmo(self, be_cosmo=None, H0=67.66, Omega_b0=0.049, Omega_dm0=0.262, Omega_k0=0.0):
        """Set the cosmology

        Parameters
        ----------
        be_cosmo: clmm.cosmology.Cosmology object, None
            Input cosmology, used if not None.
        **kwargs
            Individual cosmological parameters, see init_from_params function.
        """
        if be_cosmo:
            self._init_from_cosmo(be_cosmo)
        else:
            self.init_from_params(H0=H0, Omega_b0=Omega_b0, Omega_dm0=Omega_dm0, Omega_k0=Omega_k0)

    def get_Omega_m(self, z):
        r"""Gets the value of the dimensionless matter density

        .. math::
            \Omega_m(z) = \frac{\rho_m(z)}{\rho_\text{crit}(z)}.

        Parameters
        ----------
        z : float, array_like
            Redshift

        Returns
        -------
        Omega_m : float, numpy.ndarray
            Dimensionless matter density, :math:`\Omega_m(z)`

        Notes
        -----
        Need to decide if non-relativist neutrinos will contribute here.
        """
        if self.validate_input:
            validate_argument(locals(), "z", "float_array", argmin=0, eqmin=True)
        return self._get_Omega_m(z=z)

    def get_E2(self, z):
        r"""Gets the value of the hubble parameter (normalized at 0)

        .. math::
            E^2(z) = \frac{H(z)^{2}}{H_{0}^{2}}.

        Parameters
        ----------
        z : float
            Redshift.
        Returns
        -------
        Hubble parameter : float
            :math:`H(z)^{2}/H_{0}^{2}`.
        Notes
        -----
        Need to decide if non-relativist neutrinos will contribute here.
        """
        if self.validate_input:
            validate_argument(locals(), "z", "float_array", argmin=0, eqmin=True)
        return self._get_E2(z=z)

    def get_E2Omega_m(self, z):
        r"""Gets the value of the dimensionless matter density times the Hubble parameter squared
        (normalized at 0)

        .. math::
            \Omega_m(z) = \frac{\rho_m(z)}{\rho_\text{crit}(z)}\frac{H(z)^{2}}{H_{0}^{2}}.

        Parameters
        ----------
        z : float, array_like
            Redshift

        Returns
        -------
        Omega_m : float, numpy.ndarray
            Dimensionless matter density, :math:`\Omega_m(z)\times H(z)^{2}/H_{0}^{2}`

        Notes
        -----
        Need to decide if non-relativist neutrinos will contribute here.
        """
        if self.validate_input:
            validate_argument(locals(), "z", "float_array", argmin=0, eqmin=True)
        return self._get_E2Omega_m(z=z)

    def get_rho_m(self, z):
        r"""Gets physical matter density at a given redshift.

        Parameters
        ----------
        z : float, array_like
            Redshift

        Returns
        -------
        float, numpy.ndarray
            Matter density :math:`M_\odot\ Mpc^{-3}`
        """
        if self.validate_input:
            validate_argument(locals(), "z", "float_array", argmin=0, eqmin=True)
        return self._get_rho_m(z=z)

    def get_rho_c(self, z):
        r"""Gets physical critical density at a given redshift.

        Parameters
        ----------
        z : float
            Redshift.

        Returns
        -------
        float
            Critical density :math:`M_\odot\ Mpc^{-3}`.
        """
        if self.validate_input:
            validate_argument(locals(), "z", "float_array", argmin=0, eqmin=True)
        return self._get_rho_c(z=z)

    def eval_da_z1z2(self, z1, z2):
        r"""Computes the angular diameter distance between z1 and z2

        .. math::
            d_a(z1, z2) = \frac{c}{H_0}\frac{1}{1+z2}\int_{z1}^{z2}\frac{dz'}{E(z')}.

        Parameters
        ----------
        z1 : float, array_like
            Redshift
        z2 : float, array_like
            Redshift

        Returns
        -------
        float, numpy.ndarray
            Angular diameter distance in units :math:`M\!pc`

        Notes
        -----
        np.nan is returned for z1>z2.
        """
        if self.validate_input:
            validate_argument(locals(), "z1", "float_array", argmin=0, eqmin=True)
            validate_argument(locals(), "z2", "float_array", argmin=0, eqmin=True)
        return self._eval_da_z1z2(z1=z1, z2=z2)

    def eval_da(self, z):
        r"""Computes the angular diameter distance between 0.0 and z

        .. math::
            d_a(z) = \frac{c}{H_0}\frac{1}{1+z}\int_{0}^{z}\frac{dz'}{E(z')}.

        Parameters
        ----------
        z : float, array_like
            Redshift

        Returns
        -------
        float, numpy.ndarray
            Angular diameter distance in units :math:`M\!pc`
        """
        if self.validate_input:
            validate_argument(locals(), "z", "float_array", argmin=0, eqmin=True)
        return self._eval_da(z)

    def eval_da_a1a2(self, a1, a2=1.0):
        r"""This is a function to calculate the angular diameter distance
        between two scale factors.

        .. math::
            d_a(a1, a2) = \frac{c}{H_0}a2\int_{a2}^{a1}\frac{da'}{a'^2E(a')}

        If only a1 is specified, this function returns the angular diameter
        distance from a=1 to a1. If both a1 and a2 are specified, this function
        returns the angular diameter distance between a1 and a2.

        .. math::
            d_a(a) = \frac{c}{H_0}a\int_{a}^{1}\frac{da'}{a'^2E(a')}

        Parameters
        ----------
        a1 : float, array_like
            Scale factor
        a2 : float, array_like, optional
            Scale factor

        Returns
        -------
        float, numpy.ndarray
            Angular diameter distance in units :math:`M\!pc`
        """
        if self.validate_input:
            validate_argument(
                locals(),
                "a1",
                "float_array",
                argmin=0,
                eqmin=True,
                argmax=1,
                eqmax=True,
            )
            validate_argument(
                locals(),
                "a2",
                "float_array",
                argmin=0,
                eqmin=True,
                argmax=1,
                eqmax=True,
            )
        z1 = self.get_z_from_a(a2)
        z2 = self.get_z_from_a(a1)
        return self._eval_da_z1z2(z1, z2)

    def get_a_from_z(self, z):
        """Convert redshift to scale factor

        Parameters
        ----------
        z : float, array_like
            Redshift

        Returns
        -------
        a : float, numpy.ndarray
            Scale factor
        """
        if self.validate_input:
            validate_argument(locals(), "z", "float_array", argmin=0, eqmin=True)
        return self._get_a_from_z(z)

    def get_z_from_a(self, a):
        """Convert scale factor to redshift

        Parameters
        ----------
        a : float, array_like
            Scale factor

        Returns
        -------
        z : float, numpy.ndarray
            Redshift
        """
        if self.validate_input:
            validate_argument(
                locals(), "a", "float_array", argmin=0, eqmin=True, argmax=1, eqmax=True
            )
        return self._get_z_from_a(a)

    def rad2mpc(self, dist1, redshift):
        r"""Convert between radians and Mpc using the small angle approximation
        and :math:`d = D_A \theta`.

        Parameters
        ----------
        dist1 : float, array_like
            Input distances in radians
        redshift : float
            Redshift used to convert between angular and physical units

        Returns
        -------
        dist2 : float, numpy.ndarray
            Distances in Mpc
        """
        if self.validate_input:
            validate_argument(locals(), "dist1", "float_array", argmin=0, eqmin=True)
            validate_argument(locals(), "redshift", float, argmin=0, eqmin=True)
        return dist1 * self.eval_da(redshift)

    def mpc2rad(self, dist1, redshift):
        r"""Convert between radians and Mpc using the small angle approximation
        and :math:`d = D_A \theta`.

        Parameters
        ----------
        dist1 : float, array_like
            Input distances in Mpc
        redshift : float
            Redshift used to convert between angular and physical units

        Returns
        -------
        dist2 : float, numpy.ndarray
            Distances in radians
        """
        if self.validate_input:
            validate_argument(locals(), "dist1", "float_array", argmin=0, eqmin=True)
            validate_argument(locals(), "redshift", float, argmin=0, eqmin=True)
        return dist1 / self.eval_da(redshift)

    def eval_sigma_crit(self, z_len, z_src):
        r"""Computes the critical surface density

        Parameters
        ----------
        z_len : float
            Lens redshift
        z_src : float, array_like
            Background source galaxy redshift(s)

        Returns
        -------
        float, numpy.ndarray
            Cosmology-dependent critical surface density in units of :math:`M_\odot\ Mpc^{-2}`

        Notes
        -----
        np.inf is returned for z_src<z_len.
        """
        if self.validate_input:
            validate_argument(locals(), "z_len", float, argmin=0, eqmin=True)
            validate_argument(locals(), "z_src", "float_array", argmin=0, eqmin=True)
        return self._eval_sigma_crit(z_len=z_len, z_src=z_src)

    def eval_linear_matter_powerspectrum(self, k_vals, redshift):
        r"""Computes the linear matter power spectrum

        Parameters
        ----------
        k_vals : float, array_like
            Wavenumber k [math:`Mpc^{-1}`] values to compute the power spectrum.
        redshift : float
            Redshift to get the power spectrum.

        Returns
        -------
        float, numpy.ndarray
            Linear matter spectrum in units of math:`Mpc^{3}`
        """
        if self.validate_input:
            validate_argument(locals(), "k_vals", "float_array", argmin=0)
            validate_argument(locals(), "redshift", float, argmin=0, eqmin=True)
        return self._eval_linear_matter_powerspectrum(k_vals, redshift)



"""@file __init__.py
Theory package
"""

# ------------------------------------------------------------------------------
# Modeling backend loader
import importlib
import os
import warnings

from . import func_layer, generic
from .func_layer import (
    compute_3d_density,
    compute_convergence,
    compute_critical_surface_density_eff,
    compute_excess_surface_density,
    compute_excess_surface_density_2h,
    compute_magnification,
    compute_magnification_bias,
    compute_mean_surface_density,
    compute_reduced_tangential_shear,
    compute_surface_density,
    compute_surface_density_2h,
    compute_tangential_shear,
)
from .generic import (
    compute_magnification_bias_from_magnification,
    compute_profile_mass_in_radius,
    compute_rdelta,
    compute_reduced_shear_from_convergence,
    convert_profile_mass_concentration,
)

# Functions that do the loading of different backends


def backend_is_available(be_key):
    """Check if required backend is available"""
    conf = __backends.get(be_key)
    if not conf:
        return False
    try:
        if "preload" in conf:
            conf["preload"]()
        for module in conf["prereqs"]:
            importlib.import_module(module)
        return True
    except ImportError:
        return False


def _load_backend(be_module):
    """Loads one backend module

    Parameters
    ----------
    be_module : str
        Name of backend module
    """
    backend = importlib.import_module(f"clmm.theory.{be_module}")

    # pylint: disable=protected-access
    if all(obj in vars(backend) for obj in ("Modeling", "Cosmology")):
        # pylint: disable=global-statement
        global Modeling
        global Cosmology
        Modeling = backend.Modeling
        Cosmology = backend.Cosmology
        func_layer._modeling_object = backend.Modeling()
    else:
        func_layer._modeling_object = None


def _load_backend_fallback(be_key):
    """Loads the backend of choice if available or send a warning and try to load
    the backends in the order of the __backends dictionary.


    Parameters
    ----------
    be_key : str
        Key for the selected backend in the __backends dictionary.
    """
    first_msg = True
    for key in (be_key, *filter(lambda k: k != be_key, __backends.keys())):
        conf = __backends[key]
        if conf["available"]:
            _load_backend(conf["module"])
            return key
        if first_msg:
            warnings.warn(
                f"CLMM Backend requested '{conf['name']}' is not available, trying others..."
            )
            first_msg = False
        else:
            warnings.warn(f"* {conf['name']} BACKEND also not available")
    raise ImportError("No modeling backend available.")


def load_backend_env():
    """Loads the backend of choice if available or send a warning and try to load
    the backends in the order of the __backends dictionary.


    Parameters
    ----------
    be_key : str
        Key for the selected backend in the __backends dictionary.
    """
    #  Backend check:
    #    Checks all backends and set available to True for those that can be
    #    corretly loaded.
    for nick, be_conf in __backends.items():
        be_conf["available"] = backend_is_available(nick)

    #  Backend nick:
    #    If the environment variable CLMM_MODELING_BACKEND is set it gets its value,
    #    falls back to 'ccl' => CCL if CLMM_MODELING_BACKEND is not set.
    be_nick_env = os.environ.get("CLMM_MODELING_BACKEND", "ccl")
    if be_nick_env not in __backends:
        raise ValueError(f"CLMM Backend {be_nick_env}'' is not supported")
    #  Backend load:
    return _load_backend_fallback(be_nick_env)


##########################
# Prepare the code backend
##########################

# Global variables

# pylint: disable=invalid-name
Modeling = None
Cosmology = None

#  Preload functions:
#    Some backends depend on more complicated modules and thus on a preload
#    function.


def __numcosmo_preload():
    # pylint: disable=import-outside-toplevel
    import gi

    gi.require_version("NumCosmoMath", "1.0")
    gi.require_version("NumCosmo", "1.0")


#  Backend dictionary __backends:
#    Dictonary controling the backends, it must test if the backend is available
#    and loadable.
#    - name: The backend name;
#    - module: The actual module name, must be a .py file inside the modbackend
#      directory;
#    - prereqs: modules that need to be loadable to allow the backend to work;
#    - preload: an optional function that must be called before the modules in
#      prereqs are tested;
#    - available: must always starts False;

__backends = {
    "ccl": {"name": "ccl", "available": False, "module": "ccl", "prereqs": ["pyccl"]},
    "nc": {
        "name": "NumCosmo",
        "available": False,
        "module": "numcosmo",
        "prereqs": ["gi.repository.NumCosmoMath", "gi.repository.NumCosmo"],
        "preload": __numcosmo_preload,
    },
    "ct": {
        "name": "cluster_toolkit+astropy",
        "available": False,
        "module": "cluster_toolkit",
        "prereqs": ["cluster_toolkit", "astropy"],
    },
    "notabackend": {
        "name": "notaname",
        "available": False,
        "module": "notamodule",
        "prereqs": ["notaprereq"],
    },
}

be_nick = load_backend_env()



"""General utility functions that are used in multiple modules"""

import numpy as np


def convert_shapes_to_epsilon(shape_1, shape_2, shape_definition="epsilon", kappa=0):
    r"""Convert shape components 1 and 2 appropriately to make them estimators of the
    reduced shear once averaged.  The shape 1 and 2 components may correspond to ellipticities
    according the :math:`\epsilon`- or :math:`\chi`-definition, but also to the 1 and 2 components
    of the shear. See Bartelmann & Schneider 2001 for details
    (https://arxiv.org/pdf/astro-ph/9912508.pdf).

    The :math:`\epsilon`-ellipticity is a direct estimator of
    the reduced shear. The shear :math:`\gamma` may be converted to reduced shear :math:`g` if the
    convergence :math:`\kappa` is known. The conversions are given below.

    .. math::
     \epsilon = \frac{\chi}{1+(1-|\chi|^2)^{1/2}}

    .. math::
     g=\frac{\gamma}{1-\kappa}

    - If `shape_definition = 'chi'`, this function returns the corresponding `epsilon`
      ellipticities

    - If `shape_definition = 'shear'`, it returns the corresponding reduced shear, given the
      convergence `kappa`

    - If `shape_definition = 'epsilon'` or `'reduced_shear'`, it returns them as is as no
      conversion is needed.

    Parameters
    ----------
    shape_1 : float, numpy.ndarray
        Input shapes or shears along principal axis (g1 or e1)
    shape_2 : float, numpy.ndarray
        Input shapes or shears along secondary axis (g2 or e2)
    shape_definition : str, optional
        Definition of the input shapes, can be ellipticities 'epsilon' or 'chi' or shears 'shear'
        or 'reduced_shear'. Defaut: 'epsilon'
    kappa : float, numpy.ndarray, optional
        Convergence for transforming to a reduced shear. Default is 0

    Returns
    -------
    epsilon_1 : float, numpy.ndarray
        Epsilon ellipticity (or reduced shear) along principal axis (epsilon1)
    epsilon_2 : float, numpy.ndarray
        Epsilon ellipticity (or reduced shear) along secondary axis (epsilon2)
    """

    if shape_definition in ("epsilon", "reduced_shear"):
        epsilon_1, epsilon_2 = shape_1, shape_2
    elif shape_definition == "chi":
        chi_to_eps_conversion = 1.0 / (1.0 + (1 - (shape_1**2 + shape_2**2)) ** 0.5)
        epsilon_1, epsilon_2 = (
            shape_1 * chi_to_eps_conversion,
            shape_2 * chi_to_eps_conversion,
        )
    elif shape_definition == "shear":
        epsilon_1, epsilon_2 = shape_1 / (1.0 - kappa), shape_2 / (1.0 - kappa)
    else:
        raise TypeError("Please choose epsilon, chi, shear, reduced_shear")
    return epsilon_1, epsilon_2


def build_ellipticities(q11, q22, q12):
    """Build ellipticties from second moments. See, e.g., Schneider et al. (2006)

    Parameters
    ----------
    q11 : float, numpy.ndarray
        Second brightness moment tensor, component (1,1)
    q22 : float, numpy.ndarray
        Second brightness moment tensor, component (2,2)
    q12 :  float, numpy.ndarray
        Second brightness moment tensor, component (1,2)

    Returns
    -------
    chi1, chi2 : float, numpy.ndarray
        Ellipticities using the "chi definition"
    epsilon1, epsilon2 : float, numpy.ndarray
        Ellipticities using the "epsilon definition"
    """
    norm_x, norm_e = q11 + q22, q11 + q22 + 2 * np.sqrt(q11 * q22 - q12 * q12)
    chi1, chi2 = (q11 - q22) / norm_x, 2 * q12 / norm_x
    epsilon1, epsilon2 = (q11 - q22) / norm_e, 2 * q12 / norm_e
    return chi1, chi2, epsilon1, epsilon2


def compute_lensed_ellipticity(ellipticity1_true, ellipticity2_true, shear1, shear2, convergence):
    r"""Compute lensed ellipticities from the intrinsic ellipticities, shear and convergence.
    Following Schneider et al. (2006)

    .. math::
        \epsilon^{\text{lensed}}=\epsilon^{\text{lensed}}_1+i\epsilon^{\text{lensed}}_2=
        \frac{\epsilon^{\text{true}}+g}{1+g^\ast\epsilon^{\text{true}}},

    where, the complex reduced shear :math:`g` is obtained from the shear
    :math:`\gamma=\gamma_1+i\gamma_2` and convergence :math:`\kappa` as :math:`g =
    \gamma/(1-\kappa)`, and the complex intrinsic ellipticity is :math:`\epsilon^{\text{
    true}}=\epsilon^{\text{true}}_1+i\epsilon^{\text{true}}_2`

    Parameters
    ----------
    ellipticity1_true : float, numpy.ndarray
        Intrinsic ellipticity of the sources along the principal axis
    ellipticity2_true : float, numpy.ndarray
        Intrinsic ellipticity of the sources along the second axis
    shear1 :  float, numpy.ndarray
        Shear component (not reduced shear) along the principal axis at the source location
    shear2 :  float, numpy.ndarray
        Shear component (not reduced shear) along the 45-degree axis at the source location
    convergence :  float, numpy.ndarray
        Convergence at the source location
    Returns
    -------
    e1, e2 : float, numpy.ndarray
        Lensed ellipicity along both reference axes.
    """
    # shear (as a complex number)
    shear = shear1 + shear2 * 1j
    # intrinsic ellipticity (as a complex number)
    ellipticity_true = ellipticity1_true + ellipticity2_true * 1j
    # reduced shear
    reduced_shear = shear / (1.0 - convergence)
    # lensed ellipticity
    lensed_ellipticity = (ellipticity_true + reduced_shear) / (
        1.0 + reduced_shear.conjugate() * ellipticity_true
    )
    return np.real(lensed_ellipticity), np.imag(lensed_ellipticity)



"""General utility functions that are used in multiple modules"""

import numpy as np

from ..utils.constants import Constants as const


def arguments_consistency(arguments, names=None, prefix=""):
    r"""Make sure all arguments have the same length (or are scalars)

    Parameters
    ----------
    arguments: list, arrays, tuple
        Group of arguments to be checked
    names: list, tuple, None, optional
        Names for each array. Default: None
    prefix: str, optional
        Customized prefix for error message. Default: ''

    Returns
    -------
    list, arrays, tuple
        Group of arguments, converted to numpy arrays if they have length
    """
    sizes = [len(arg) if hasattr(arg, "__len__") else None for arg in arguments]
    # check there is a name for each argument
    if names:
        if len(names) != len(arguments):
            raise TypeError(
                f"names (len={len(names)}) must have same length "
                f"as arguments (len={len(arguments)})"
            )
        msg = ", ".join([f"{n}({s})" for n, s in zip(names, sizes)])
    else:
        msg = ", ".join([f"{s}" for s in sizes])
    # check consistency
    if any(sizes):
        # Check that all of the inputs have length and they match
        if not all(sizes) or any(s != sizes[0] for s in sizes[1:]):
            # make error message
            raise TypeError(f"{prefix} inconsistent sizes: {msg}")
        return tuple(np.array(arg) for arg in arguments)
    return arguments


def _patch_rho_crit_to_cd2018(rho_crit_external):
    r"""Convertion factor for rho_crit of any external modult to
    CODATA 2018+IAU 2015

    rho_crit_external: float
        Critical density of the Universe in units of :math:`M_\odot\ Mpc^{-3}`
    """

    rhocrit_mks = 3.0 * 100.0 * 100.0 / (8.0 * np.pi * const.GNEWT.value)
    rhocrit_cd2018 = (
        rhocrit_mks * 1000.0 * 1000.0 * const.PC_TO_METER.value * 1.0e6 / const.SOLAR_MASS.value
    )

    return rhocrit_cd2018 / rho_crit_external


_valid_types = {
    float: (float, int, np.floating, np.integer),
    int: (int, np.integer),
    "float_array": (float, int, np.floating, np.integer),
    "int_array": (int, np.integer),
    "array": (list, tuple, np.ndarray),
}


def _is_valid(arg, valid_type):
    r"""Check if argument is of valid type, supports arrays.

    Parameters
    ----------
    arg: any
        Argument to be tested.
    valid_type: str, type
        Valid types for argument, options are object types, list/tuple of types, or:

            * 'int_array' - interger, interger array
            * 'float_array' - float, float array

    Returns
    -------
    valid: bool
        Is argument valid
    """
    if valid_type == "function":
        return callable(arg)
    return (
        isinstance(arg[0], _valid_types[valid_type])
        if (valid_type in ("int_array", "float_array") and np.iterable(arg))
        else isinstance(arg, _valid_types.get(valid_type, valid_type))
    )


def validate_argument(
    loc,
    argname,
    valid_type,
    none_ok=False,
    argmin=None,
    argmax=None,
    eqmin=False,
    eqmax=False,
    shape=None,
):
    r"""Validate argument type and raise errors.

    Parameters
    ----------
    loc: dict
        Dictionary with all input arguments. Should be locals().
    argname: str
        Name of argument to be tested.
    valid_type: str, type
        Valid types for argument, options are object types, list/tuple of types, or:

            * 'int_array' - integer, integer array
            * 'float_array' - float, float array

    none_ok: bool, optional
        If True, accepts None as a valid type. Default: False
    argmin : int, float, None, optional
        Minimum value allowed. Default: None
    argmax : int, float, None, optional
        Maximum value allowed. Default: None
    eqmin : bool, optional
        If True, accepts min(arg)==argmin. Default: False
    eqmax : bool, optional
        If True, accepts max(arg)==argmax. Default: False
    shape : tuple of ints, None, optional
        Shape of object allowed. Default: None
    """
    var = loc[argname]
    # Check for None
    if none_ok and (var is None):
        return
    # Check for type
    valid = (
        any(_is_valid(var, types) for types in valid_type)
        if isinstance(valid_type, (list, tuple))
        else _is_valid(var, valid_type)
    )
    if not valid:
        err = f"{argname} must be {valid_type}, received {type(var).__name__}"
        raise TypeError(err)
    # Check min/max
    if any(t is not None for t in (argmin, argmax)):
        try:
            var_array = np.array(var, dtype=float)
        except Exception as exc:
            err = (
                f"{argname} ({type(var).__name__}) cannot be converted to number"
                " for min/max validation."
            )
            raise TypeError(err) from exc
        if argmin is not None:
            if var_array.min() < argmin if eqmin else var_array.min() <= argmin:
                err = (
                    f"{argname} must be greater than {argmin},"
                    f" received min({argname}): {var_array.min()}"
                )
                raise ValueError(err)
        if argmax is not None:
            if var_array.max() > argmax if eqmax else var_array.max() >= argmax:
                err = (
                    f"{argname} must be lesser than {argmax},"
                    f" received max({argname}): {var_array.max()}"
                )
                raise ValueError(err)
    # Check for shape
    if shape is not None:
        if np.shape(var) != shape:
            err = (
                f"{argname} must be of shape {shape}," f"received shape({argname}): {np.shape(var)}"
            )
            raise ValueError(err)


def _validate_ra(loc, ra_name, is_array):
    r"""Validate RA type and raise errors.

    Parameters
    ----------
    loc: dict
        Dictionary with all input arguments. Should be locals().
    ra_name: str
        Name of RA in args.
    is_array: bool
        Accepts array as input.
    """
    v_type = "float_array" if is_array else (float, str)
    validate_argument(loc, ra_name, v_type, argmin=-360, eqmin=True, argmax=360, eqmax=True)


def _validate_dec(loc, dec_name, is_array):
    r"""Validate DEC type and raise errors.

    Parameters
    ----------
    loc: dict
        Dictionary with all input arguments. Should be locals().
    dec_name: str
        Name of DEC in args.
    is_array: bool
        Accepts array as input.
    """
    v_type = "float_array" if is_array else (float, str)
    validate_argument(loc, dec_name, v_type, argmin=-90, eqmin=True, argmax=90, eqmax=True)


def _validate_is_deltasigma_sigma_c(is_deltasigma, sigma_c):
    r""" "Validate the compatibility between is_deltasigma and sigma_c arguments.


    Parameters
    ----------
    is_deltasigma: bool
        If `False`, values are computed for shear, else they are computed for :math:`\Delta \Sigma`.
    sigma_c : None, array_like
        Critical (effective) surface density in units of :math:`M_\odot\ Mpc^{-2}`.
    """
    if is_deltasigma and sigma_c is None:
        raise TypeError("sigma_c (=None) must be provided when is_deltasigma=True")
    if not is_deltasigma and sigma_c is not None:
        raise TypeError(f"sigma_c (={sigma_c}) must be None when is_deltasigma=False")


def _validate_coordinate_system(loc, argname):
    r"""Validate the coordinate system.

    Parameters
    ----------
    loc: dict
        Dictionary with all input arguments. Should be locals().
    argname: str
        Name of argument to be tested.
    """
    validate_argument(loc, argname, str)
    if loc[argname] not in ["celestial", "euclidean"]:
        raise ValueError(f"{argname} must be 'celestial' or 'euclidean'.")


class DiffArray:
    """Array where arr1==arr2 is actually all(arr1==arr)"""

    def __init__(self, array):
        self.value = np.array(array)

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        if type(other) != type(self):
            return False
        if self.value.size != other.value.size:
            return False
        return (self.value == other.value).all()

    def __repr__(self):
        out = str(self.value)
        if self.value.size > 4:
            out = self._get_lim_str(out) + " ... " + self._get_lim_str(out[::-1])[::-1]
        return out

    def _get_lim_str(self, out):
        # pylint: disable=undefined-loop-variable
        # get count starting point
        for init_index, char in enumerate(out):
            if all(char != _char for _char in "[]() "):
                break
        # get str
        sep = 0
        for i, char in enumerate(out[init_index + 1 :]):
            sep += int(char == " " and out[i + init_index] != " ")
            if sep == 2:
                break
        return out[: i + init_index + 1]



"""@file cluster_toolkit.py
Modeling using cluster_toolkit
"""

# Functions to model halo profiles
import cluster_toolkit as ct
import numpy as np

from ..cosmology.cluster_toolkit import AstroPyCosmology
from ..utils import _patch_rho_crit_to_cd2018
from .parent_class import CLMModeling


def _assert_correct_type_ct(arg):
    """Convert the argument to a type compatible with cluster_toolkit
    cluster_toolkit does not handle correctly scalar arguments that are
    not float or numpy array and others that contain non-float64 elements.
    It only convert lists to the correct type. To circumvent this we
    pre-convert all arguments going to cluster_toolkit to the appropriated
    types.

    Parameters
    ----------
    arg : array_like or scalar

    Returns
    -------
    scale_factor : numpy.ndarray
        Scale factor
    """
    if np.isscalar(arg):
        return float(arg)
    return np.array(arg).astype(np.float64, order="C", copy=False)


class CTCLMModeling(CLMModeling):
    r"""Object with functions for halo mass modeling

    Attributes
    ----------
    backend: str
        Name of the backend being used
    massdef : str
        Profile mass definition ("mean", "critical" - letter case independent)
    delta_mdef : int
        Mass overdensity definition.
    halo_profile_model : str
        Profile model parameterization ("nfw" - letter case independent)
    cosmo: Cosmology
        Cosmology object
    hdpm: Object
        Backend object with halo profiles
    mdef_dict: dict
        Dictionary with the definitions for mass
    hdpm_dict: dict
        Dictionary with the definitions for profile
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=abstract-method

    def __init__(
        self,
        massdef="mean",
        delta_mdef=200,
        halo_profile_model="nfw",
        validate_input=True,
    ):
        CLMModeling.__init__(self, validate_input)
        # Update class attributes
        self.backend = "ct"
        self.hdpm_dict = {"nfw": "nfw"}
        self.cosmo_class = AstroPyCosmology
        # Attributes exclusive to this class
        self.cor_factor = _patch_rho_crit_to_cd2018(2.77533742639e11)
        self.__mdelta = 0.0  # internal use only
        self.__cdelta = 0.0  # internal use only

        # Set halo profile and cosmology
        self.set_cosmo(None)
        # Exploit cluster-toolkit's calculation to get massdef='critical'
        # CT takes the argument 'Omega_m' to culculate rho_m(z) from rho_c (in M_sun*h^2/Mpc^3)
        # Passing (H(z)/H0)^2*Omega_m(z)=E(z)^2*Omega_m(z) is expected
        # Passing (H(z)/H0)^2=E^2 will effectively change the massdef to 'critical'
        # by calculating rho_c(z), instead of rho_m(z), from rho_c
        self.mdef_dict = {
            "mean": self.cosmo.get_E2Omega_m,
            "critical": self.cosmo.get_E2,
        }

        self.set_halo_density_profile(halo_profile_model, massdef, delta_mdef)

    # Functions implemented by child class

    def _update_halo_density_profile(self):
        """updates halo density profile with set internal properties"""
        # pylint: disable=unnecessary-pass
        pass

    def _get_concentration(self):
        """get concentration"""
        return self.__cdelta

    def _get_mass(self):
        """get mass"""
        return self.__mdelta

    def _set_concentration(self, cdelta):
        """set concentration"""
        self.__cdelta = cdelta

    def _set_mass(self, mdelta):
        """set mass"""
        self.__mdelta = mdelta

    def _eval_3d_density(self, r3d, z_cl):
        """eval 3d density"""
        h = self.cosmo["h"]
        Omega_m = self.mdef_dict[self.massdef](z_cl) * self.cor_factor
        return (
            ct.density.rho_nfw_at_r(
                _assert_correct_type_ct(r3d) * h,
                self.mdelta * h,
                self.cdelta,
                Omega_m,
                delta=self.delta_mdef,
            )
            * h**2
        )

    def _eval_surface_density(self, r_proj, z_cl):
        """eval surface density"""
        h = self.cosmo["h"]
        Omega_m = self.mdef_dict[self.massdef](z_cl) * self.cor_factor
        return (
            ct.deltasigma.Sigma_nfw_at_R(
                _assert_correct_type_ct(r_proj) * h,
                self.mdelta * h,
                self.cdelta,
                Omega_m,
                delta=self.delta_mdef,
            )
            * h
            * 1.0e12
        )  # pc**-2 to Mpc**-2

    def _eval_mean_surface_density(self, r_proj, z_cl):
        r"""Computes the mean value of surface density inside radius `r_proj`

        Parameters
        ----------
        r_proj : array_like
            Projected radial position from the cluster center in :math:`M\!pc`.
        z_cl: float
            Redshift of the cluster

        Returns
        -------
        numpy.ndarray, float
            Excess surface density in units of :math:`M_\odot\ Mpc^{-2}`.

        Note
        ----
        This function just adds eval_surface_density+eval_excess_surface_density
        """
        return self.eval_surface_density(r_proj, z_cl) + self.eval_excess_surface_density(
            r_proj, z_cl
        )

    def _eval_excess_surface_density(self, r_proj, z_cl):
        """eval excess surface density"""
        if np.min(r_proj) < 1.0e-11:
            raise ValueError(
                f"Rmin = {np.min(r_proj):.2e} Mpc!"
                " This value is too small and may cause computational issues."
            )

        h = self.cosmo["h"]
        Omega_m = self.mdef_dict[self.massdef](z_cl) * self.cor_factor

        r_proj = _assert_correct_type_ct(r_proj) * h
        # Computing sigma on a larger range than the radial range requested,
        # with at least 1000 points.
        sigma_r_proj = np.logspace(
            np.log10(np.min(r_proj)) - 1,
            np.log10(np.max(r_proj)) + 1,
            np.max([1000, 10 * np.array(r_proj).size]),
        )
        sigma = self.eval_surface_density(sigma_r_proj / h, z_cl) / (h * 1e12)  # rm norm for ct
        # ^ Note: Let's not use this naming convention when transfering ct to ccl....
        return (
            ct.deltasigma.DeltaSigma_at_R(
                r_proj,
                sigma_r_proj,
                sigma,
                self.mdelta * h,
                self.cdelta,
                Omega_m,
                delta=self.delta_mdef,
            )
            * h
            * 1.0e12
        )  # pc**-2 to Mpc**-2


Cosmology = AstroPyCosmology
Modeling = CTCLMModeling
